"""Generate XML Schema Definition (XSD) corresponding to the meta-model."""

import re

# noinspection PyUnresolvedReferences
import xml.dom.minidom
import xml.etree.ElementTree as ET
from typing import TextIO, MutableMapping, Optional, Tuple, List, Any, Mapping, Final

import greenery
from icontract import ensure, require

import aas_core_codegen.xsd
from aas_core_codegen import (
    naming,
    specific_implementations,
    intermediate,
    run,
    infer_for_schema,
)
from aas_core_codegen.common import Error, assert_never, Identifier
from aas_core_codegen.parse import retree as parse_retree
from aas_core_codegen.xsd import naming as xsd_naming

assert aas_core_codegen.xsd.__doc__ == __doc__


def _define_for_enumeration(enumeration: intermediate.Enumeration) -> List[ET.Element]:
    """
    Generate the definitions for an ``enumeration``.
    The root element is to be *extended* with the resulting list.
    """
    restriction = ET.Element("xs:restriction", {"base": "xs:string"})
    for literal in enumeration.literals:
        restriction.append(ET.Element("xs:enumeration", {"value": literal.value}))

    element = ET.Element(
        "xs:simpleType", {"name": xsd_naming.type_name(enumeration.name)}
    )
    element.append(restriction)

    return [element]


_PRIMITIVE_MAP = {
    intermediate.PrimitiveType.BOOL: "xs:boolean",
    intermediate.PrimitiveType.INT: "xs:long",
    intermediate.PrimitiveType.FLOAT: "xs:double",
    intermediate.PrimitiveType.STR: "xs:string",
    intermediate.PrimitiveType.BYTEARRAY: "xs:base64Binary",
}
assert all(literal in _PRIMITIVE_MAP for literal in intermediate.PrimitiveType)

# noinspection RegExpSimplifiable
_ESCAPE_BACKSLASH_X_RE = re.compile(r"\\x([a-fA-f0-9]{2})")


def _undo_escaping_backslash_x_in_pattern(pattern: str) -> str:
    """
    Undo the escaping of `\\x??` in the ``pattern``.

    This is necessary since XML Schema Validators do not know how to handle such escape
    sequences in the patterns and need the verbatim characters.
    """
    parts = []  # type: List[str]
    cursor = None  # type: Optional[int]
    for mtch in re.finditer(_ESCAPE_BACKSLASH_X_RE, pattern):
        if cursor is None:
            parts.append(pattern[: mtch.start()])
        else:
            parts.append(pattern[cursor : mtch.start()])

        ascii_code = int(mtch.group(1), base=16)
        character = chr(ascii_code)
        parts.append(character)
        cursor = mtch.end()

    if cursor is None:
        parts.append(pattern)
    else:
        if cursor < len(pattern):
            parts.append(pattern[cursor:])

    return "".join(parts)


# noinspection RegExpSimplifiable
_ESCAPE_BACKSLASH_X_U_U_RE = re.compile(
    r"(\\x([a-fA-f0-9]{2})|\\u([a-fA-f0-9]{4})|\\U([a-fA-f0-9]{8}))"
)


def _undo_escaping_backslash_x_u_and_U_in_pattern(pattern: str) -> str:
    """
    Undo the escaping of ``\\x??``, ``\\u????`` and ``\\U????????`` in the ``pattern``.

    This is necessary since Greenery does not know how to handle such escape
    sequences in the patterns and need the verbatim characters.
    """
    parts = []  # type: List[str]
    cursor = None  # type: Optional[int]
    for mtch in re.finditer(_ESCAPE_BACKSLASH_X_U_U_RE, pattern):
        if cursor is None:
            parts.append(pattern[: mtch.start()])
        else:
            parts.append(pattern[cursor : mtch.start()])

        substring = mtch.group(0)
        assert len(substring) > 2
        assert substring[0] == "\\"

        hex_code = substring[2:]
        code_point = int(hex_code, base=16)
        character = chr(code_point)
        parts.append(character)
        cursor = mtch.end()

    if cursor is None:
        parts.append(pattern)
    else:
        if cursor < len(pattern):
            parts.append(pattern[cursor:])

    return "".join(parts)


class _AnchorRemover(parse_retree.PassThroughVisitor):
    """
    Remove anchors from a regex in-place.

    We need to remove the anchors (``^``, ``$``) since patterns in the XSD are always
    anchored.

    This is necessary since otherwise the schema validation fails.
    See: https://stackoverflow.com/questions/4367914/regular-expression-in-xml-schema-definition-fails
    """

    def visit_concatenation(self, node: parse_retree.Concatenation) -> None:
        """Visit the ``concatenation``."""
        new_concatenants = []  # type: List[parse_retree.Term]
        for concatenant in node.concatenants:
            if not (
                isinstance(concatenant.value, parse_retree.Symbol)
                and concatenant.value.kind
                in (parse_retree.SymbolKind.START, parse_retree.SymbolKind.END)
            ):
                new_concatenants.append(concatenant)

        node.concatenants = new_concatenants
        for concatenant in new_concatenants:
            self.visit(concatenant)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _translate_pattern(pattern: str) -> Tuple[Optional[str], Optional[str]]:
    """Translate the pattern to obtain the equivalent in XSD."""
    pattern = _undo_escaping_backslash_x_in_pattern(pattern)

    parsed, error = parse_retree.parse(values=[pattern])
    if error is not None:
        regex_line, pointer_line = parse_retree.render_pointer(error.cursor)
        return None, f"{error.message}\n{regex_line}\n{pointer_line}"
    assert parsed is not None

    remover = _AnchorRemover()
    remover.visit(parsed)

    values = parse_retree.render(regex=parsed)
    parts = []  # type: List[str]
    for value in values:
        assert isinstance(value, str), (
            "Only strings expected when rendering a pattern "
            "supplied originally as a string"
        )
        parts.append(value)

    return "".join(parts), None


class _SimpleTypeRestriction:
    """Model the restrictions applied on a simple type."""

    pattern: Final[Optional[str]]

    min_length: Final[Optional[int]]
    max_length: Final[Optional[int]]

    @require(
        lambda pattern, min_length, max_length: (pattern is not None)
        or (min_length is not None)
        or (max_length is not None),
        "At least one constraint defined",
    )
    def __init__(
        self,
        pattern: Optional[str] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
    ) -> None:
        self.pattern = pattern
        self.min_length = min_length
        self.max_length = max_length


class _SimpleType:
    """
    Model a simple type in XSD.

    This is meant to be represented either with ``<xs:simpleType>``, if there are
    restrictions, or as an attribute ``type="..."`` in the containing element, when
    there are no restrictions.
    """

    tajp: Final[str]
    restriction: Final[Optional[_SimpleTypeRestriction]]

    def __init__(
        self, tajp: str, restriction: Optional[_SimpleTypeRestriction]
    ) -> None:
        self.tajp = tajp
        self.restriction = restriction


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _translate_to_simple_type(
    primitive_type: intermediate.PrimitiveType,
    constraints: Optional[infer_for_schema.Constraints],
) -> Tuple[Optional[_SimpleType], Optional[str]]:
    restriction: Optional[_SimpleTypeRestriction] = None

    min_length: Optional[int] = None
    max_length: Optional[int] = None
    pattern: Optional[str] = None

    if constraints is not None:
        if constraints.len_constraint is not None:
            min_length = constraints.len_constraint.min_value
            max_length = constraints.len_constraint.max_value

        if constraints.patterns is not None:
            # NOTE (mristin):
            # We skip the general XML character pattern. It makes greenery
            # unbearably slow since it instantiates *each* character in
            # the character range. Since XML engines can not deal with the special
            # characters any ways, there is no need to include this constraint in
            # the XSD pattern restrictions.
            patterns_relevant_for_xsd = [
                pattern_constraint
                for pattern_constraint in constraints.patterns
                if pattern_constraint.pattern
                != (
                    "^[\\x09\\x0A\\x0D\\x20-\\uD7FF\\uE000-\\uFFFD"
                    "\\U00010000-\\U0010FFFF]*$"
                )
            ]

            if len(patterns_relevant_for_xsd) > 0:
                translated_pattern: Optional[str]

                if len(patterns_relevant_for_xsd) == 1:
                    translated_pattern, error = _translate_pattern(
                        patterns_relevant_for_xsd[0].pattern
                    )
                    if error is not None:
                        return None, error
                else:
                    # NOTE (mristin, 2023-02-27):
                    # The module ``greenery`` is not annotated with types at the moment.
                    merger = None  # type: Optional[Any]
                    for pattern_constraint in patterns_relevant_for_xsd:
                        # NOTE (mristin, 2023-02-27):
                        # Greenery expects the characters to be in Unicode and not escaped.
                        translated_for_greenery = (
                            _undo_escaping_backslash_x_u_and_U_in_pattern(
                                pattern_constraint.pattern
                            )
                        )

                        try:
                            parsed = greenery.parse(translated_for_greenery)
                        except Exception as exception:
                            if translated_for_greenery == pattern_constraint.pattern:
                                return None, (
                                    f"The greenery failed to parse "
                                    f"the pattern {translated_for_greenery!r}: "
                                    f"{exception}"
                                )
                            else:
                                return None, (
                                    f"The greenery failed to parse "
                                    f"the pattern {translated_for_greenery!r} "
                                    f"(which was originally "
                                    f"{pattern_constraint.pattern!r}): {exception}"
                                )

                        if merger is None:
                            merger = parsed
                        else:
                            merger = merger & parsed

                    assert merger is not None

                    translated_pattern, error = _translate_pattern(str(merger))
                    if error is not None:
                        return None, error

                assert translated_pattern is not None
                pattern = translated_pattern

    if (min_length is not None) or (max_length is not None) or (pattern is not None):
        restriction = _SimpleTypeRestriction(
            min_length=min_length, max_length=max_length, pattern=pattern
        )

    return (
        _SimpleType(tajp=_PRIMITIVE_MAP[primitive_type], restriction=restriction),
        None,
    )


class _TypeElementOrTypeIdentifier:
    """
    Represent a rendering of a type annotation.

    A type annotation can either be represented as a ``type`` attribute of an element,
    or as a proper ``xs:simpleType`` or ``xs:complexType`` element.

    For example:

    .. code-block:: xml

        <xs:element name="something" type="xs:string" />

    or:

    .. code-block:: xml

        <xs:element name="someLanguage">
            <xs:simpleType>
                <xs:restriction base="xs:string">
                    <xs:minLength value="1" />
                </xs:restriction>
            </xs:simpleType>
        </xs:element>
    """

    tajp: Final[Optional[str]]

    element: Final[Optional[ET.Element]]

    @require(lambda tajp, element: (tajp is not None) ^ (element is not None))
    def __init__(
        self, tajp: Optional[str] = None, element: Optional[ET.Element] = None
    ) -> None:
        self.tajp = tajp
        self.element = element


def _value_to_type_element_or_type_identifier(
    type_annotation: intermediate.TypeAnnotationExceptOptional,
    constraints_by_value: infer_for_schema.ConstraintsByValue,
) -> Tuple[Optional[_TypeElementOrTypeIdentifier], Optional[str]]:
    """
    Translate the given type annotation to XSD.

    Return the either type element or type identifier, or an error, if any.
    """
    if isinstance(type_annotation, intermediate.OurTypeAnnotation):
        if type_annotation.our_type.name == "Value_data_type":
            # NOTE (mristin, 2022-11-10):
            # Please see :py:const:`_EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE`
            # for the explanation why we hard-wire the ``Value_data_type`` here.
            return _TypeElementOrTypeIdentifier(tajp="valueDataType"), None

    primitive_type = intermediate.try_primitive_type(type_annotation)

    if primitive_type is not None:
        simple_type, simple_type_error = _translate_to_simple_type(
            primitive_type=primitive_type,
            constraints=constraints_by_value.get(type_annotation, None),
        )
        if simple_type_error is not None:
            return None, (
                f"Failed to translate the type annotation {type_annotation} "
                f"to xs:simpleType: {simple_type_error}"
            )

        assert simple_type is not None

        if simple_type.restriction is None:
            return _TypeElementOrTypeIdentifier(tajp=simple_type.tajp), None
        else:
            xs_simple_type = ET.Element("xs:simpleType")

            xs_restriction = ET.SubElement(
                xs_simple_type, "xs:restriction", {"base": simple_type.tajp}
            )

            # NOTE (mristin):
            # We define the pattern first for the legacy reasons -- already published
            # XSD schemas defined the pattern first, so we want to keep the diffs
            # minimal even though it intuitively makes more sense to start with
            # length.

            if simple_type.restriction.pattern is not None:
                ET.SubElement(
                    xs_restriction,
                    "xs:pattern",
                    {"value": simple_type.restriction.pattern},
                )

            if simple_type.restriction is not None:
                if simple_type.restriction.min_length is not None:
                    ET.SubElement(
                        xs_restriction,
                        "xs:minLength",
                        {"value": str(simple_type.restriction.min_length)},
                    )

                if simple_type.restriction.max_length is not None:
                    ET.SubElement(
                        xs_restriction,
                        "xs:maxLength",
                        {"value": str(simple_type.restriction.max_length)},
                    )

            return _TypeElementOrTypeIdentifier(element=xs_simple_type), None
    else:
        if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
            raise AssertionError("This execution path must have been handled before.")

        elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
            our_type = type_annotation.our_type

            if isinstance(our_type, intermediate.Enumeration):
                return (
                    _TypeElementOrTypeIdentifier(
                        tajp=xsd_naming.type_name(our_type.name)
                    ),
                    None,
                )

            elif isinstance(our_type, intermediate.ConstrainedPrimitive):
                raise AssertionError(
                    "This execution path must have been handled before."
                )

            elif isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            ):
                # NOTE (mristin):
                # We generate choices only if there are at least one concrete descendant.
                # Otherwise, the choice is not generated. Hence, we need to reference
                # a choice only if there is actually one.
                #
                # This is especially necessary for abstract classes with no descendants
                # which we still want to include in the schema. We simply generate an empty
                # element in the schema for such abstract classes without descendants.

                if len(our_type.concrete_descendants) > 0:
                    xs_complex_type = ET.Element("xs:complexType")
                    xs_sequence = ET.SubElement(xs_complex_type, "xs:sequence")

                    ET.SubElement(
                        xs_sequence,
                        "xs:group",
                        {"ref": xsd_naming.choice_group_name(our_type.name)},
                    )

                    return (
                        _TypeElementOrTypeIdentifier(
                            element=xs_complex_type,
                        ),
                        None,
                    )

                else:
                    return (
                        _TypeElementOrTypeIdentifier(
                            tajp=xsd_naming.type_name(our_type.name)
                        ),
                        None,
                    )
            else:
                # noinspection PyTypeChecker
                assert_never(our_type)

        elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
            assert not isinstance(
                type_annotation.items, intermediate.OptionalTypeAnnotation
            ), (
                "(mristin, 2026-05-08): Only lists of non-optionals are supported "
                "at the moment. If you see this, please contact the developers."
            )

            xs_complex_type = ET.Element("xs:complexType")
            xs_sequence = ET.SubElement(xs_complex_type, "xs:sequence")

            item_element: ET.Element

            if isinstance(
                type_annotation.items, intermediate.OurTypeAnnotation
            ) and isinstance(
                type_annotation.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ):
                # NOTE (mristin):
                # All the other cases introduce ``<v>`` element to capture the items,
                # but lists of classes use the name of the class, so we have to handle
                # it here differently.

                # NOTE (mristin):
                # We generate choices only if there are at least one concrete descendant.
                # Otherwise, the choice is not generated. Hence, we need to reference
                # a choice only if there is actually one.
                #
                # This is especially necessary for abstract classes with no descendants
                # which we still want to include in the schema. We simply generate an empty
                # element in the schema for such abstract classes without descendants.

                if len(type_annotation.items.our_type.concrete_descendants) > 0:
                    item_element = ET.Element(
                        "xs:group",
                        {
                            "ref": xsd_naming.choice_group_name(
                                type_annotation.items.our_type.name
                            )
                        },
                    )
                else:
                    item_element = ET.Element(
                        "xs:element",
                        {
                            "name": naming.xml_class_name(
                                type_annotation.items.our_type.name
                            ),
                            "type": xsd_naming.type_name(
                                type_annotation.items.our_type.name
                            ),
                        },
                    )
            else:
                (
                    items_type_element_or_identifier,
                    translation_error,
                ) = _value_to_type_element_or_type_identifier(
                    type_annotation=type_annotation.items,
                    constraints_by_value=constraints_by_value,
                )

                if translation_error is not None:
                    return None, (
                        f"Failed to translate the type annotation {type_annotation} to "
                        f"a type element or a type identifier: {translation_error}"
                    )

                assert items_type_element_or_identifier is not None

                item_element = ET.Element("xs:element", {"name": "v"})

                if items_type_element_or_identifier.tajp is not None:
                    item_element.attrib["type"] = items_type_element_or_identifier.tajp

                else:
                    assert items_type_element_or_identifier.element is not None

                    item_element.append(items_type_element_or_identifier.element)

            xs_sequence.append(item_element)

            min_occurs = "0"
            max_occurs = "unbounded"

            constraints = constraints_by_value.get(type_annotation, None)
            if constraints is not None and constraints.len_constraint is not None:
                if constraints.len_constraint.min_value is not None:
                    min_occurs = str(constraints.len_constraint.min_value)

                if constraints.len_constraint.max_value is not None:
                    max_occurs = str(constraints.len_constraint.max_value)

            item_element.attrib["minOccurs"] = min_occurs
            item_element.attrib["maxOccurs"] = max_occurs

            return _TypeElementOrTypeIdentifier(element=xs_complex_type), None

        else:
            # noinspection PyTypeChecker
            assert_never(type_annotation)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_properties(
    cls: intermediate.ClassUnion,
    constraints_by_class: Mapping[
        intermediate.ClassUnion, infer_for_schema.ConstraintsByValue
    ],
) -> Tuple[Optional[List[ET.Element]], Optional[List[Error]]]:
    """Define the properties of the ``cls`` as a sequence of tags."""
    sequence = []  # type: List[ET.Element]
    errors = []  # type: List[Error]

    constraints_by_value = constraints_by_class[cls]

    for prop in cls.properties:
        # NOTE (mristin):
        # While we allow our classes to tighten the constraints from the parents, XSD
        # does not allow us to *easily* tighten the constraints in a group.
        #
        # Instead of making it really complicated, we simply ignore the tightening
        # of the constraints in the children classes.
        #
        # If we ever want to implement this feature, we have to define three intermediate
        # structures:
        # 1) Abstract parent class,
        # 2) Restricted class, only tightening the elements with
        #    ``<xs:restriction base="...">``, and
        # 3) Extended class which includes properties specified only for the class with
        #    ``<xs:extension base="...">``.

        if prop.specified_for is not cls:
            continue

        type_anno = intermediate.beneath_optional(prop.type_annotation)

        xs_element = ET.Element("xs:element", {"name": naming.xml_property(prop.name)})

        # fmt: off
        (
            type_element_or_identifier,
            type_error
        ) = _value_to_type_element_or_type_identifier(
            type_annotation=type_anno,
            constraints_by_value=constraints_by_value
        )
        # fmt: on

        if type_error is not None:
            errors.append(
                Error(
                    prop.parsed.node,
                    f"Failed to translate the type annotation, "
                    f"{type_anno}, of the property {prop.name} "
                    f"to XSD: {type_error}",
                )
            )
            continue

        assert type_element_or_identifier is not None

        if type_element_or_identifier.tajp is not None:
            xs_element.attrib["type"] = type_element_or_identifier.tajp
        else:
            assert type_element_or_identifier.element is not None

            xs_element.append(type_element_or_identifier.element)

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            xs_element.attrib["minOccurs"] = "0"
            xs_element.attrib["maxOccurs"] = "1"

        sequence.append(xs_element)

    if len(errors) > 0:
        return None, errors

    return sequence, None


def _generate_xs_group_for_class(
    cls: intermediate.ClassUnion,
    constraints_by_class: Mapping[
        intermediate.ClassUnion, infer_for_schema.ConstraintsByValue
    ],
) -> Tuple[Optional[ET.Element], Optional[Error]]:
    """Generate the ``xs:group`` representation of the class properties."""
    properties, properties_errors = _define_properties(
        cls=cls, constraints_by_class=constraints_by_class
    )

    if properties_errors is not None:
        return None, Error(
            cls.parsed.node,
            f"Failed to generate xs:group for the class {cls.name!r}",
            properties_errors,
        )

    assert properties is not None

    xs_sequence = ET.Element("xs:sequence")
    for inheritance in cls.inheritances:
        inheritance_xs_group = ET.Element(
            "xs:group", {"ref": xsd_naming.group_name(inheritance.name)}
        )
        xs_sequence.append(inheritance_xs_group)

    xs_sequence.extend(properties)

    xs_group = ET.Element("xs:group", {"name": xsd_naming.group_name(cls.name)})
    xs_group.append(xs_sequence)

    return xs_group, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_for_class(
    cls: intermediate.ClassUnion,
    constraints_by_class: Mapping[
        intermediate.ClassUnion, infer_for_schema.ConstraintsByValue
    ],
) -> Tuple[Optional[List[ET.Element]], Optional[Error]]:
    """
    Generate the definitions for the class ``cls``.

    The root element is to be *extended* with the resulting list.
    """
    # NOTE (mristin, 2022-03-30):
    # We define each set of properties in a group. Then we reference these groups
    # among the complex types.
    # See: https://stackoverflow.com/questions/1198755/xml-schemas-with-multiple-inheritance

    xs_group, xs_group_error = _generate_xs_group_for_class(
        cls=cls, constraints_by_class=constraints_by_class
    )
    if xs_group_error is not None:
        return None, xs_group_error

    assert xs_group is not None

    xs_group_ref = ET.Element("xs:group", {"ref": xsd_naming.group_name(cls.name)})

    xs_sequence = ET.Element("xs:sequence")
    xs_sequence.append(xs_group_ref)

    complex_type = ET.Element(
        "xs:complexType", {"name": xsd_naming.type_name(cls.name)}
    )
    complex_type.append(xs_sequence)

    return [xs_group, complex_type], None


@require(lambda cls: len(cls.concrete_descendants) > 0)
def _generate_choice_group(cls: intermediate.ClassUnion) -> ET.Element:
    """Generate a group that defines a choice of concrete descendants."""
    xs_choice = ET.Element("xs:choice")

    if isinstance(cls, intermediate.ConcreteClass):
        xs_choice.append(
            ET.Element(
                "xs:element",
                {
                    "name": naming.xml_class_name(cls.name),
                    "type": xsd_naming.type_name(cls.name),
                },
            )
        )

    for descendant in cls.concrete_descendants:
        xs_choice.append(
            ET.Element(
                "xs:element",
                {
                    "name": naming.xml_class_name(descendant.name),
                    "type": xsd_naming.type_name(descendant.name),
                },
            )
        )

    xs_group = ET.Element("xs:group", {"name": xsd_naming.choice_group_name(cls.name)})
    xs_group.append(xs_choice)
    return xs_group


_WHITESPACE_RE = re.compile(r"\s+")


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _retrieve_implementation_specific_elements(
    cls: intermediate.ClassUnion,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[List[ET.Element]], Optional[List[Error]]]:
    """Parse the elements from the implementation-specific snippet."""
    implementation_key = specific_implementations.ImplementationKey(f"{cls.name}.xml")

    text = spec_impls.get(implementation_key, None)
    if text is None:
        return None, [
            Error(
                cls.parsed.node,
                f"The implementation is missing "
                f"for the implementation-specific class: {implementation_key}",
            )
        ]

    implementation_root: ET.Element

    try:
        implementation_root = ET.fromstring(text)
    except Exception as err:
        return None, [
            Error(
                cls.parsed.node,
                f"Failed to parse the XML out of "
                f"the specific implementation {implementation_key}: {err}",
            )
        ]

    errors = []  # type: List[Error]
    for descendant in implementation_root.iter():
        if descendant.text is not None and not _WHITESPACE_RE.fullmatch(
            descendant.text
        ):
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Unexpected text "
                    f"in the specific implementation {implementation_key} "
                    f"in an element with tag {descendant.tag!r}: {descendant.text!r}",
                )
            )
            continue

        if descendant.tail is not None and not _WHITESPACE_RE.fullmatch(
            descendant.tail
        ):
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Unexpected tail text "
                    f"in the specific implementation {implementation_key} "
                    f"in an element with tag {descendant.tag!r}: {descendant.tail!r}",
                )
            )
            continue

    if len(errors) > 0:
        return None, errors

    # Ignore the implementation root since it defines a partial schema
    elements = []  # type: List[ET.Element]
    for child in implementation_root:
        elements.append(child)

    return elements, None


def _sort_by_tags_and_names_in_place(root: ET.Element) -> None:
    """
    Sort the children elements by tag and name attribute in place.

    This makes diffing and searching in the schema a bit easier.
    """
    groups = []  # type: List[ET.Element]
    simple_types = []  # type: List[ET.Element]
    complex_types = []  # type: List[ET.Element]
    miscellaneous = []  # type: List[ET.Element]
    elements = []  # type: List[ET.Element]

    for child in root:
        if child.tag == "xs:group":
            groups.append(child)
        elif child.tag == "xs:simpleType":
            simple_types.append(child)
        elif child.tag == "xs:complexType":
            complex_types.append(child)
        elif child.tag == "xs:element":
            elements.append(child)
        else:
            miscellaneous.append(child)

    for element_list in [groups, simple_types, complex_types, miscellaneous, elements]:
        element_list.sort(key=lambda elt: elt.attrib.get("name", ""))

    children = groups + simple_types + complex_types + elements + miscellaneous

    assert len(children) == len(root)
    root[:] = children


_EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE = (
    "(mristin, 2022-09-02) "
    'We provide an internal data type ``valueDataType`` to correspond to "any XSD '
    'atomic type as specified via DataTypeDefXsd". We need this type since we '
    "hard-wire ``Value_data_type`` to it. We could have made "
    "the class ``Value_data_type``implementation-specific and defined its "
    "representation manually as a snippet, including ``valueDataType``.\n\n"
    "However, we decided against that. This would be a major hurdle for "
    "other code and test data generators (which can treat ``Value_data_type`` "
    "simply as string). Therefore, we make the XSD generator "
    "a bit more hacky instead of complicating the other generators.\n\n"
    "If in the future, for whatever reason, the semantic of ``Value_data_type`` "
    "changes (or the type is renamed), be careful to maintain backwards "
    "compatibility here! You probably want to distinguish different versions "
    "of the meta-model and act accordingly. At that point, it might also make "
    "sense to refactor this schema generator to a separate repository, and "
    "fix it to a particular range of meta-model versions."
)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the XML Schema Definition (XSD) based on the ``symbol_table."""
    root_element_key = specific_implementations.ImplementationKey("root_element.xml")

    root_element_as_text = spec_impls.get(root_element_key, None)
    if root_element_as_text is None:
        return None, [
            Error(
                None,
                f"The implementation snippet for the root element "
                f"is missing: {root_element_key}",
            )
        ]

    root: ET.Element
    try:
        root = ET.fromstring(root_element_as_text)
    except ET.ParseError as err:
        return None, [
            Error(
                None, f"Failed to parse the root element from {root_element_key}: {err}"
            )
        ]

    # NOTE (mristin, 2022-03-30):
    # We need to use minidom to extract the ``xmlns`` property as ElementTree removes
    # it.
    # noinspection PyUnresolvedReferences
    minidom_doc = xml.dom.minidom.parseString(root_element_as_text)

    if not minidom_doc.documentElement.hasAttribute("xmlns"):
        return None, [
            Error(
                None,
                f"The implementation snippet for the root element "
                f"is missing the 'xmlns' attribute: {root_element_key}",
            )
        ]

    xmlns = minidom_doc.documentElement.getAttribute("xmlns")

    if xmlns != symbol_table.meta_model.xml_namespace:
        return None, [
            Error(
                None,
                f"The 'xmlns' attribute of the implementation snippet "
                f"{root_element_key} for the root element "
                f"and the '__xml_namespace__' of the meta-model "
                f"do not coincide: "
                f"{xmlns!r} != {symbol_table.meta_model.xml_namespace!r}",
            )
        ]

    if not minidom_doc.documentElement.hasAttribute("targetNamespace"):
        return None, [
            Error(
                None,
                f"The implementation snippet for the root element "
                f"is missing the 'targetNamespace' attribute: {root_element_key}",
            )
        ]

    target_namespace = minidom_doc.documentElement.getAttribute("targetNamespace")
    if target_namespace != symbol_table.meta_model.xml_namespace:
        return None, [
            Error(
                None,
                f"The 'targetNamespace' attribute of the implementation snippet "
                f"{root_element_key} for the root element "
                f"and the '__xml_namespace__' of the meta-model "
                f"do not coincide: "
                f"{target_namespace!r} != {symbol_table.meta_model.xml_namespace!r}",
            )
        ]

    assert root is not None

    errors = []  # type: List[Error]

    # NOTE (mristin, 2022-04-09):
    # We remove any whitespace tail and text in all the tags, and make sure there is no
    # unexpected text anywhere.
    for element in root.iter():
        if element.text is not None:
            if _WHITESPACE_RE.fullmatch(element.text):
                element.text = None
            else:
                errors.append(
                    Error(
                        None,
                        f"Unexpected text in an element with tag {element.tag!r} "
                        f"from the snippet {root_element_key!r}: {element.text!r}",
                    )
                )

        if element.tail is not None:
            if _WHITESPACE_RE.fullmatch(element.tail):
                element.tail = None
            else:
                errors.append(
                    Error(
                        None,
                        f"Unexpected tail in an element with tag {element.tag!r} "
                        f"from the snippet {root_element_key!r}: {element.tail!r}",
                    )
                )

    if len(errors) > 0:
        return None, errors

    constraints_by_class, some_errors = infer_for_schema.infer_constraints_by_class(
        symbol_table=symbol_table
    )

    if some_errors is not None:
        errors.extend(some_errors)

    # NOTE (mristin, 2022-11-10):
    # Please see :py:const:`_EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE` for
    # the explanation why we retrieve the ``Value_data_type`` here
    value_data_type_cls = symbol_table.find_our_type(Identifier("Value_data_type"))

    if value_data_type_cls is None:
        errors.append(
            Error(
                None,
                "XSD generator expected to find our type ``Value_data_type``, but it "
                "was not present in the meta-model.\n\n"
                + _EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE,
            )
        )
    elif not isinstance(value_data_type_cls, intermediate.ConstrainedPrimitive):
        errors.append(
            Error(
                None,
                "XSD generator expected ``Value_data_type`` to be "
                "a constrained primitive,  but got: {type(value_data_type_cls)}.\n\n"
                + _EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE,
            )
        )
    elif value_data_type_cls.constrainee != intermediate.PrimitiveType.STR:
        errors.append(
            Error(
                None,
                f"XSD generator expected ``Value_data_type`` to be a constrained "
                f"primitive of strings, "
                f"but got: {value_data_type_cls.constrainee}.\n\n"
                + _EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE,
            )
        )
    else:
        # Our type ``Value_data_type`` is as expected.
        pass

    if len(errors) > 0:
        return None, errors

    assert constraints_by_class is not None

    ids_of_our_types_in_properties = (
        intermediate.collect_ids_of_our_types_in_properties(symbol_table=symbol_table)
    )

    # region Specify ``valueDataType``

    assert value_data_type_cls is not None

    value_data_type_element = ET.Element(
        "xs:simpleType", attrib={"name": "valueDataType"}
    )

    value_data_type_element.append(
        ET.Element(
            "xs:restriction",
            attrib={"base": "xs:string"},
        )
    )

    root.append(value_data_type_element)

    # endregion

    for our_type in symbol_table.our_types:
        if our_type.name == "Value_data_type":
            # NOTE (mristin, 2022-11-10):
            # Please see :py:const:`_EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE`
            # for the explanation why we hard-wire the ``Value_data_type`` here
            continue

        elements: Optional[List[ET.Element]]

        if (
            isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            )
            and our_type.is_implementation_specific
        ):
            elements, impl_spec_errors = _retrieve_implementation_specific_elements(
                cls=our_type, spec_impls=spec_impls
            )
            if impl_spec_errors is not None:
                errors.extend(impl_spec_errors)
                continue

            assert elements is not None
        else:
            if isinstance(our_type, intermediate.Enumeration):
                if id(our_type) not in ids_of_our_types_in_properties:
                    continue

                elements = _define_for_enumeration(enumeration=our_type)

            elif isinstance(our_type, intermediate.ConstrainedPrimitive):
                # NOTE (mristin, 2022-03-30):
                # We in-line the constraints from the constrained primitives directly
                # in the properties. We do not want to introduce separate definitions
                # for them as that would make it more difficult for downstream code
                # generators to generate meaningful code.

                continue

            elif isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            ):
                elements, definition_error = _define_for_class(
                    cls=our_type, constraints_by_class=constraints_by_class
                )

                if definition_error is not None:
                    errors.append(definition_error)
                    continue

                assert elements is not None

                if len(our_type.concrete_descendants) > 0:
                    choice_group = _generate_choice_group(cls=our_type)
                    elements.append(choice_group)
            else:
                # noinspection PyTypeChecker
                assert_never(our_type)

        assert elements is not None
        root.extend(elements)

    if len(errors) > 0:
        return None, errors

    # Tag name 🠒 (name 🠒 element)
    observed_definitions = dict(
        dict()
    )  # type: MutableMapping[str, MutableMapping[str, ET.Element]]

    for element in root:
        name = element.attrib.get("name", None)
        if name is None:
            continue

        observed_for_tag = observed_definitions.get(element.tag, None)
        if observed_for_tag is None:
            observed_for_tag = dict()
            observed_definitions[element.tag] = observed_for_tag

        observed_element = observed_for_tag.get(name, None)
        if observed_element is not None:
            ours = ET.tostring(element, encoding="unicode", method="xml")
            theirs = ET.tostring(observed_element, encoding="unicode", method="xml")

            errors.append(
                Error(
                    None,
                    f"There are conflicting definitions in the schema "
                    f"with the name {name!r}:\n"
                    f"\n"
                    f"{ours}\n"
                    f"\n"
                    f"and\n"
                    f"\n"
                    f"{theirs}",
                )
            )
        else:
            observed_for_tag[name] = element

    if len(errors) > 0:
        return None, errors

    _sort_by_tags_and_names_in_place(root)

    # NOTE (mristin, 2022-03-30):
    # For some unknown reason, ElementTree erases the xmlns property of the root
    # element. Therefore, we need to add it here manually.
    root.attrib["xmlns"] = xmlns

    text = ET.tostring(root, encoding="unicode", method="xml")

    # NOTE (mristin, 2021-11-23):
    # This approach is slow, but effective. As long as the meta-model is not too big,
    # this should work.
    # noinspection PyUnresolvedReferences
    pretty_text = xml.dom.minidom.parseString(text).toprettyxml(indent="  ")

    return pretty_text, None


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the code."""
    code, errors = _generate(
        symbol_table=context.symbol_table, spec_impls=context.spec_impls
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the XML Schema Definition "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "schema.xsd"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the XML Schema Definition to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0



"""Generate code of the index file which import all the other modules."""

import io
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen import specific_implementations
from aas_core_codegen.common import (
    Stripped,
    Error,
)
from aas_core_codegen.typescript import (
    common as typescript_common,
    description as typescript_description,
)


# fmt: off
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate code of the index file which import all the other modules."""
    key = specific_implementations.ImplementationKey("package_documentation.txt")

    text = spec_impls.get(key, None)
    if text is None:
        return None, [
            Error(
                None,
                f"The package documentation snippet is missing "
                f"in the specific implementations: {key}",
            )
        ]

    text = Stripped(
        f"""\
{text}

@packageDocumentation"""
    )

    comment = Stripped(typescript_description.documentation_comment(text))

    blocks = [
        comment,
        typescript_common.WARNING,
        Stripped(
            """\
export * as common from "./common";
export * as constants from "./constants";
export * as jsonization from "./jsonization";
export * as xmlization from "./xmlization";
export * as stringification from "./stringification";
export * as types from "./types";
export * as verification from "./verification";"""
        ),
        typescript_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the verification."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)


def _generate_for_cls(cls: intermediate.ConcreteClass) -> List[Stripped]:
    """Generate the tests for a class."""
    model_type_literal = golang_common.string_literal(naming.json_model_type(cls.name))

    deserialization_function = golang_naming.function_name(
        Identifier(f"{cls.name}_from_jsonable")
    )

    blocks = []  # type: List[Stripped]

    test_name = golang_naming.function_name(Identifier(f"Test_{cls.name}_OK"))

    blocks.append(
        Stripped(
            f"""\
func {test_name}(t *testing.T) {{
{I}pths := aastesting.FindFilesBySuffixRecursively(
{II}filepath.Join(
{III}aastesting.TestDataDir,
{III}"Json",
{III}"Expected",
{III}{model_type_literal},
{II}),
{II}".json",
{I})
{I}sort.Strings(pths)

{I}for _, pth := range pths {{
{II}jsonable := aastesting.MustReadJsonable(
{III}pth,
{II})

{II}deserialized, deseriaErr := aasjsonization.{deserialization_function}(
{III}jsonable,
{II})
{II}if deseriaErr != nil {{
{III}t.Fatalf(
{IIII}"Unexpected deserialization error from %s: %s",
{IIII}pth, deseriaErr.Error(),
{III})
{III}return
{II}}}

{II}var errors []*aasverification.VerificationError
{II}aasverification.Verify(
{III}deserialized,
{III}func(veriErr *aasverification.VerificationError) (abort bool) {{
{IIII}errors = append(errors, veriErr)
{IIII}return
{III}}},
{II})

{II}ok := assertNoVerificationErrors(
{III}t,
{III}deserialized,
{III}pth,
{II})
{II}if !ok {{
{III}return
{II}}}
{I}}}
}}"""
        )
    )

    test_name = golang_naming.function_name(Identifier(f"Test_{cls.name}_fail"))

    blocks.append(
        Stripped(
            f"""\
func {test_name}(t *testing.T) {{
{I}pattern := filepath.Join(
{II}aastesting.TestDataDir,
{II}"Json",
{II}"Unexpected",
{II}"Invalid",
{II}"*",  // This asterisk represents the cause.
{II}{model_type_literal},
{I})

{I}causeDirs, err := filepath.Glob(pattern)
{I}if err != nil {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Failed to find cause directories matching %s: %s",
{IIII}pattern, err.Error(),
{III}),
{II})
{I}}}

{I}for _, causeDir := range causeDirs {{
{II}pths := aastesting.FindFilesBySuffixRecursively(
{III}causeDir,
{III}".json",
{II})
{II}sort.Strings(pths)

{II}for _, pth := range pths {{
{III}jsonable := aastesting.MustReadJsonable(
{IIII}pth,
{III})

{III}relPth, err := filepath.Rel(aastesting.TestDataDir, pth)
{III}if err != nil {{
{IIII}panic(
{IIIII}fmt.Sprintf(
{IIIIII}"Failed to compute the relative path of %s to %s: %s",
{IIIIII}aastesting.TestDataDir, pth, err.Error(),
{IIIII}),
{IIII})
{III}}}

{III}expectedPth := filepath.Join(
{IIII}aastesting.TestDataDir,
{IIII}"VerificationError",
{IIII}filepath.Dir(relPth),
{IIII}filepath.Base(relPth)+".errors",
{III})

{III}deserialized, deseriaErr := aasjsonization.{deserialization_function}(
{IIII}jsonable,
{III})
{III}if deseriaErr != nil {{
{IIII}t.Fatalf(
{IIIII}"Unexpected deserialization error from %s: %s",
{IIIII}pth, deseriaErr.Error(),
{IIII})
{IIII}return
{III}}}

{III}var errors []*aasverification.VerificationError
{III}aasverification.Verify(
{IIII}deserialized,
{IIII}func(err *aasverification.VerificationError) (abort bool) {{
{IIIII}errors = append(errors, err)
{IIIII}return
{IIII}}},
{III})

{III}ok := assertEqualsExpectedOrRerecordVerificationErrors(
{IIII}t,
{IIII}errors,
{IIII}pth,
{IIII}expectedPth,
{III})
{III}if !ok {{
{IIII}return
{III}}}
{II}}}
{I}}}
}}"""
        )
    )

    return blocks


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate code to test the verification."""
    blocks = [
        Stripped("package verification_test"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"encoding/json"
{I}"fmt"
{I}"os"
{I}"path/filepath"
{I}"sort"
{I}"strings"
{I}"testing"
{I}aasjsonization "{repo_url}/jsonization"
{I}aastesting "{repo_url}/aastesting"
{I}aastypes "{repo_url}/types"
{I}aasverification "{repo_url}/verification"
)"""
        ),
        Stripped(
            f"""\
// Assert that there are no verification errors in the `instance` de-serialized
// from `source`.
func assertNoVerificationErrors(
{I}t *testing.T,
{I}instance aastypes.IClass,
{I}source string,
) (ok bool) {{
{I}errors := make([]*aasverification.VerificationError, 0)
{I}aasverification.Verify(
{II}instance,
{II}func(err *aasverification.VerificationError) (abort bool) {{
{III}errors = append(errors, err)
{III}return
{II}}},
{I})

{I}ok = true
{I}if len(errors) > 0 {{
{II}ok = false

{II}var sb strings.Builder

{II}sb.WriteString(
{III}fmt.Sprintf(
{IIII}"Expected no errors when verifying the instance de-serialized from "+
{IIIII}"%s, but got %d error(s)\\n",
{IIII}source, len(errors),
{III}),
{II})

{II}for i, err := range errors {{
{III}sb.WriteString(
{IIII}fmt.Sprintf(
{IIIII}"Error %d:\\n%s: %s\\n",
{IIIII}i+1,
{IIIII}err.PathString(),
{IIIII}err.Message,
{IIII}),
{III})
{II}}}

{II}jsonable, seriaErr := aasjsonization.ToJsonable(instance)
{II}if seriaErr != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to serialize instance to JSON obtained from %s: %s",
{IIIII}source, seriaErr.Error(),
{IIII}),
{III})
{II}}}
{II}jsonableBytes, err := json.Marshal(jsonable)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to marshal to JSON an instance serialized from %s: %s",
{IIIII}source, err.Error(),
{IIII}),
{III})
{II}}}

{II}sb.WriteString("Instance:\\n")
{II}sb.WriteString(string(jsonableBytes))

{II}t.Fatal(sb.String())
{I}}}
{I}return
}}"""
        ),
        Stripped(
            f"""\
// Assert that either the verification errors match the recorded ones at `pth`, if
// [aastesting.RecordMode] is set, or re-record the verification errors at `pth`.
func assertEqualsExpectedOrRerecordVerificationErrors(
{I}t *testing.T,
{I}errors []*aasverification.VerificationError,
{I}source string,
{I}expectedPth string,
) (ok bool) {{
{I}ok = true
{I}if len(errors) == 0 {{
{II}ok = false
{II}t.Fatalf(
{III}"Expected at least one verification error, "+
{IIII}"but got none when verifying the model loaded from: %s",
{III}source,
{II})
{I}}}

{I}parts := make([]string, len(errors))
{I}for i, verErr := range errors {{
{II}parts[i] = fmt.Sprintf(
{III}"%s: %s",
{III}verErr.PathString(),
{III}verErr.Message,
{II})
{I}}}

{I}// Add a newline for POSIX systems
{I}got := strings.Replace(strings.Join(parts, ";\\n"), "\\r\\n", "\\n", -1) + "\\n"

{I}if aastesting.RecordMode {{
{II}parent := filepath.Dir(expectedPth)
{II}err := os.MkdirAll(parent, os.ModePerm)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to create the directory %s: %s", parent, err.Error(),
{IIII}),
{III})
{II}}}

{II}err = os.WriteFile(expectedPth, []byte(got), 0644)
{II}if err != nil {{
{III}ok = false
{III}t.Fatalf("Failed to write to %s: %s", expectedPth, err.Error())
{III}return
{II}}}
{I}}} else {{
{II}_, err := os.Stat(expectedPth)
{II}if err != nil {{
{III}ok = false
{III}t.Fatalf(
{IIII}"Failed to stat the file %s: %s; if the file does not exist, "+
{IIIII}"you probably want to record the test data by "+
{IIIII}"setting the environment variable %s",
{IIII}expectedPth, err.Error(), aastesting.RecordModeEnvironmentVariableName,
{III})
{III}return
{II}}}

{II}var bb []byte
{II}bb, err = os.ReadFile(expectedPth)
{II}if err != nil {{
{III}ok = false
{III}t.Fatalf("Failed to read from %s: %s", expectedPth, err.Error())
{III}return
{II}}}
{II}expected := strings.ReplaceAll(string(bb), "\\r\\n", "\\n")

{II}if expected != got {{
{III}ok = false
{III}t.Fatalf(
{IIII}"The expected verification errors (read from %s) in the model "+
{IIIII}"de-serialized from %s do not match the obtained ones. "+
{IIIII}"Expected:\\n%s\\nGot:\\n%s",
{IIII}expectedPth, source, expected, got,
{III})
{III}return
{II}}}
{I}}}
{I}return
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        blocks.extend(_generate_for_cls(cls=concrete_cls))

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the XML de/serialization of interfaces."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the XML de/serialization of interfaces."""
    blocks = [
        Stripped(
            """\
/**
 * Test XML de/serialization of interfaces.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as AasXmlization from "../src/xmlization";
import * as AasTypes from "../src/types";

import * as TestCommonXmlization from "./commonXmlization";"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.classes:
        if cls.interface is None or len(cls.interface.implementers) == 0:
            continue

        interface_name_typescript = typescript_naming.interface_name(cls.name)
        as_function = typescript_naming.function_name(Identifier(f"as_{cls.name}"))

        for implementer_cls in cls.interface.implementers:
            if (
                implementer_cls.serialization is None
                or not implementer_cls.serialization.with_model_type
            ):
                continue

            cls_name_typescript = typescript_naming.class_name(implementer_cls.name)
            load_minimal_name = typescript_naming.function_name(
                Identifier(f"load_minimal_{implementer_cls.name}")
            )

            blocks.append(
                Stripped(
                    f"""\
test(
{I}"{interface_name_typescript} XML round-trip " +
{I}"starting from {cls_name_typescript} OK",
{I}() => {{
{II}const instance = TestCommonXmlization.{load_minimal_name}();

{II}const xmlText = AasXmlization.toXmlString(instance);
{II}const anotherInstanceOrError = AasXmlization.fromXmlString(xmlText);
{II}expect(anotherInstanceOrError.error).toBeNull();
{II}const anotherInstance = anotherInstanceOrError.mustValue();

{II}const asInterface = AasTypes.{as_function}(anotherInstance);
{II}expect(asInterface).not.toBeNull();
{I}}}
);"""
                )
            )

        blocks.append(
            Stripped(
                f"""\
test("{interface_name_typescript} XML deserialization fail", () => {{
{I}const instanceOrError = AasXmlization.fromXmlString("This is not XML.");
{I}expect(instanceOrError.error).not.toBeNull();
}});"""
            )
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the JSON de/serialization of enumerations."""

import io
from typing import List

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


# fmt: off
@require(
    lambda enumeration: len(enumeration.literals) > 0,
    "Enumeration must have at least one literal as enumerations without literals "
    "can not be tested"
)
# fmt: on
def _generate_round_trip_test_for_enum(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate the test that the de-serialization equals the serialization."""
    test_name = golang_naming.function_name(
        Identifier(f"test_{enumeration.name}_round_trip_OK")
    )

    literals_joined = "\n".join(
        f"{golang_common.string_literal(literal.value)},"
        for literal in enumeration.literals
    )

    deserialization_function = golang_naming.function_name(
        Identifier(f"{enumeration.name}_from_jsonable")
    )

    serialization_function = golang_naming.function_name(
        Identifier(f"{enumeration.name}_to_jsonable")
    )

    return Stripped(
        f"""\
func {test_name}(t *testing.T) {{
{I}literals := []string{{
{II}{indent_but_first_line(literals_joined, II)}
{I}}}

{I}for _, literal := range literals {{
{II}source := fmt.Sprintf("<string literal %s>", literal)
{II}jsonable := any(literal)

{II}deserialized, deseriaErr := aasjsonization.{deserialization_function}(
{III}jsonable,
{II})
{II}ok := assertNoDeserializationError(t, deseriaErr, source)
{II}if !ok {{
{III}return
{II}}}

{II}anotherJsonable, seriaErr :=
{III}aasjsonization.{serialization_function}(deserialized)
{II}ok = assertNoSerializationError(t, seriaErr, source)
{II}if !ok {{
{III}return
{II}}}

{II}ok = assertSerializationEqualsDeserialization(
{III}t,
{III}jsonable,
{III}anotherJsonable,
{III}source,
{II})
{II}if !ok {{
{III}return
{II}}}
{I}}}
}}"""
    )


def _generate_deserialization_fail_for_enum(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate the test for de-serialization of an invalid value."""
    test_name = golang_naming.function_name(
        Identifier(f"test_{enumeration.name}_deserialization_fail")
    )

    deserialization_function = golang_naming.function_name(
        Identifier(f"{enumeration.name}_from_jsonable")
    )

    enum_name = golang_naming.enum_name(enumeration.name)

    invalid_literal = "THIS-CANNOT-POSSIBLY-BE-VALID"
    trial = 0
    while invalid_literal in enumeration.literal_value_set:
        trial += 1
        infix = "-EVER" * trial
        invalid_literal = f"THIS-CANNOT{infix}-POSSIBLY-BE-VALID"

    return Stripped(
        f"""\
func {test_name}(t *testing.T) {{
{I}jsonable := any({golang_common.string_literal(invalid_literal)})

{I}_, err := aasjsonization.{deserialization_function}(
{II}jsonable,
{I})

{I}if err == nil {{
{II}t.Fatal("Expected a deserialization error, but got none.")
{II}return
{I}}}

{I}deseriaErr, ok := err.(*aasjsonization.DeserializationError)
{I}if !ok {{
{II}t.Fatalf("Expected a de-serialization error, but got: %v", err)
{II}return
{I}}}

{I}pathString := deseriaErr.PathString()
{I}if len(pathString) != 0 {{
{II}t.Fatalf(
{III}"Expected an empty path in error, but got: %s",
{III}pathString,
{II})
{II}return
{I}}}

{I}expectedMessage :=
{II}"Expected a string representation of {enum_name}, " +
{II}"but got {invalid_literal}"

{I}if deseriaErr.Message != expectedMessage {{
{II}t.Fatalf(
{III}"Expected the deserialization error:\\n%s\\n, but got:\\n%s",
{III}expectedMessage, deseriaErr.Message,
{II})
{II}return
{I}}}
}}"""
    )


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate code to test the JSON de/serialization of enumerations."""
    blocks = [
        Stripped("package jsonization_test"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}"testing"
{I}aasjsonization "{repo_url}/jsonization"
)"""
        ),
    ]  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        blocks.append(_generate_round_trip_test_for_enum(enumeration=enumeration))
        blocks.append(_generate_deserialization_fail_for_enum(enumeration=enumeration))

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the ``descend`` methods and ``PassThroughVisitor``."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the ``descend`` methods and ``PassThroughVisitor``."""
    blocks = [
        Stripped(
            """\
/**
 * Test `descend*` functions and `PassThroughVisitor`.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as path from "path";
import * as fs from "fs";

import * as AasTypes from "../src/types";
import * as TestCommon from "./common";
import * as TestCommonJsonization from "./commonJsonization";"""
        ),
        Stripped(
            f"""\
/**
 * Record a human-readable trace of a descent over an object tree.
 */
class TracingVisitor extends AasTypes.PassThroughVisitor {{
{I}readonly log = new Array<string>();

{I}visit(that: AasTypes.Class) {{
{II}this.log.push(TestCommon.traceMark(that));
{II}super.visit(that);
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Expect that the trace from {{@link types.Class.descend}} and
 * {{@link TracingVisitor}} are equal.
 *
 * @param instance - to be descended through
 * @throws an {{@link Error}} if the two traces are not equal
 */
function expectDescendAndPassThroughVisitorSame(
{I}instance: AasTypes.Class
): void {{
{I}const traceFromDescend = new Array<string>();
{I}for (const subInstance of instance.descend()) {{
{II}traceFromDescend.push(TestCommon.traceMark(subInstance));
{I}}}

{I}const visitor = new TracingVisitor();
{I}visitor.visit(instance);
{I}const traceFromVisitor = visitor.log;

{I}expect(traceFromVisitor.length).toBeGreaterThan(0);
{I}expect(TestCommon.traceMark(instance)).toStrictEqual(traceFromVisitor[0]);

{I}traceFromVisitor.shift();

{I}expect(traceFromVisitor).toStrictEqual(traceFromDescend);
}}"""
        ),
        Stripped(
            f"""\
/**
 * Compare the trace against the golden one from the test data,
 * or re-record the trace if {{@link common.RECORD_MODE}}.
 *
 * @param instance - to be traced
 * @param expectedPath - path to the golden trace
 */
function compareOrRecordTrace(
{I}instance: AasTypes.Class,
{I}expectedPath: string
) {{
{I}const lines = new Array<string>();
{I}for (const descendant of instance.descend()) {{
{II}lines.push(TestCommon.traceMark(descendant));
{I}}}
{I}// NOTE (mristin, 2022-12-09):
{I}// We add a new line for POSIX systems which prefer a new line
{I}// at the end of the file.
{I}lines.push("");
{I}const got = lines.join("\\n");

{I}if (TestCommon.RECORD_MODE) {{
{II}const parent = path.dirname(expectedPath);
{II}if (!fs.existsSync(parent)) {{
{III}fs.mkdirSync(parent, {{recursive: true}});
{II}}}
{II}fs.writeFileSync(expectedPath, got, "utf-8");
{I}}} else {{
{II}if (!fs.existsSync(expectedPath)) {{
{III}throw new Error(
{IIII}`The file with the recorded trace does not exist: ${{expectedPath}}; ` +
{IIII}`you probably want to set the environment ` +
{IIII}`variable ${{TestCommon.RECORD_MODE_ENVIRONMENT_VARIABLE_NAME}}?`
{III});
{II}}}

{II}const expected =
{III}fs.readFileSync(expectedPath, "utf-8")
{III}.replace(/\\r\\n/g, "\\n");
{II}expect(got).toStrictEqual(expected);
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_typescript = typescript_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        load_maximal_name = typescript_naming.function_name(
            Identifier(f"load_maximal_{concrete_cls.name}")
        )

        blocks.append(
            Stripped(
                f"""\
test("descend of {cls_name_typescript}", () => {{
{I}const instance = TestCommonJsonization.{load_maximal_name}();

{I}compareOrRecordTrace(
{II}instance,
{II}path.join(
{III}TestCommon.TEST_DATA_DIR,
{III}"descend",
{III}{typescript_common.string_literal(cls_name_json)},
{III}"maximal.json.trace"
{II})
{I});
}});"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
test("descend against PassThroughVisitor", () => {{
{I}const instance = TestCommonJsonization.{load_maximal_name}();

{I}expectDescendAndPassThroughVisitorSame(instance);
}});"""
            )
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Merge constrained primitives as property constraints."""
import collections
import itertools
from typing import (
    Tuple,
    Optional,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    Set,
    Union,
    Iterator,
)

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Error, assert_never
from aas_core_codegen.infer_for_schema import (
    _len as infer_for_schema_len,
    _pattern as infer_for_schema_pattern,
    _set as infer_for_schema_set,
)
from aas_core_codegen.infer_for_schema._types import (
    Constraints,
    LenConstraint,
    PatternConstraint,
    SetOfPrimitivesConstraint,
    SetOfEnumerationLiteralsConstraint,
    ConstraintsByValue,
    MutableConstraintsByValue,
)


def _min_or_none(that: Optional[int], other: Optional[int]) -> Optional[int]:
    """Compute the minimum or return None if both values are None."""
    if that is not None and other is not None:
        return min(that, other)

    elif that is None and other is not None:
        return other

    elif that is not None and other is None:
        return that

    elif that is None and other is None:
        return None

    else:
        raise AssertionError("Unhandled execution path")


def _max_or_none(that: Optional[int], other: Optional[int]) -> Optional[int]:
    """Compute the maximum or return None if both values are None."""
    if that is not None and other is not None:
        return max(that, other)

    elif that is None and other is not None:
        return other

    elif that is not None and other is None:
        return that

    elif that is None and other is None:
        return None

    else:
        raise AssertionError("Unhandled execution path")


def _merge_len_constraints(
    that: Optional[LenConstraint], other: Optional[LenConstraint]
) -> Optional[LenConstraint]:
    if that is not None and other is not None:
        return LenConstraint(
            min_value=_max_or_none(that.min_value, other.min_value),
            max_value=_min_or_none(that.max_value, other.max_value),
        )

    elif that is not None and other is None:
        return that

    elif that is None and other is not None:
        return other

    elif that is None and other is None:
        return None

    else:
        raise AssertionError("Unhandled execution path")


def _merge_pattern_constraints(
    that: Optional[Sequence[PatternConstraint]],
    other: Optional[Sequence[PatternConstraint]],
) -> Optional[Sequence[PatternConstraint]]:
    if that is not None and other is not None:
        observed_pattern_set: Set[str] = set()

        result: List[PatternConstraint] = []

        for pattern_constraint in itertools.chain(that, other):
            if pattern_constraint.pattern not in observed_pattern_set:
                result.append(pattern_constraint)
                observed_pattern_set.add(pattern_constraint.pattern)

        return result

    elif that is not None and other is None:
        return that

    elif that is None and other is not None:
        return other

    elif that is None and other is None:
        return None

    else:
        raise AssertionError("Unhandled execution path")


def _merge_set_of_primitives_constraints(
    that: Optional[SetOfPrimitivesConstraint],
    other: Optional[SetOfPrimitivesConstraint],
) -> Optional[SetOfPrimitivesConstraint]:
    if that is not None and other is not None:
        if that.a_type != other.a_type:
            raise ValueError(
                "Constraints on sets of primitives of different primitive types "
                f"can not be merged together; that primitive type is {that.a_type}, "
                f"other primitive type is {other.a_type}."
            )

        value_histo: MutableMapping[
            Union[bool, int, float, str, bytearray], int
        ] = collections.OrderedDict()

        literal_by_value: MutableMapping[
            Union[bool, int, float, str, bytearray], intermediate.PrimitiveSetLiteral
        ] = {}

        for literal in itertools.chain(that.literals, other.literals):
            if literal.value not in value_histo:
                value_histo[literal.value] = 1
            else:
                value_histo[literal.value] += 1

            literal_by_value[literal.value] = literal

        return SetOfPrimitivesConstraint(
            # NOTE (mristin):
            # We simply pick one of the types as we assert before that that.a_type
            # and other.a_type must be the same.
            a_type=that.a_type,
            # NOTE (mristin):
            # We need to compute the intersection, so we only pick the literals which we
            # observed twice, once in each list of literals.
            literals=[
                literal_by_value[value]
                for value, count in value_histo.items()
                if count == 2
            ],
        )

    elif that is not None and other is None:
        return that

    elif that is None and other is not None:
        return other

    elif that is None and other is None:
        return None

    else:
        raise AssertionError("Unhandled execution path")


def _merge_set_of_enumeration_literals_constraints(
    that: Optional[SetOfEnumerationLiteralsConstraint],
    other: Optional[SetOfEnumerationLiteralsConstraint],
) -> Optional[SetOfEnumerationLiteralsConstraint]:
    if that is not None and other is not None:
        if that.enumeration is not other.enumeration:
            raise ValueError(
                "Constraints on sets of enumeration literals of different enumerations "
                f"can not be merged together; that enumeration is {that.enumeration}, "
                f"other enumeration is {other.enumeration}."
            )

        literal_by_id: MutableMapping[int, intermediate.EnumerationLiteral] = dict()

        literal_id_histo: MutableMapping[int, int] = collections.OrderedDict()

        for literal in itertools.chain(that.literals, other.literals):
            literal_id = id(literal)

            if literal_id not in literal_id_histo:
                literal_id_histo[literal_id] = 1
            else:
                literal_id_histo[literal_id] += 1

            literal_by_id[literal_id] = literal

        return SetOfEnumerationLiteralsConstraint(
            # NOTE (mristin):
            # We simply pick one of the enumerations as we assert before that
            # enumeration and other enumeration are one and the same.
            enumeration=that.enumeration,
            # NOTE (mristin):
            # We need to compute the intersection, so we only pick the literals which we
            # observed twice, once in each list of literals.
            literals=[
                literal_by_id[literal_id]
                for literal_id, count in literal_id_histo.items()
                if count == 2
            ],
        )

    elif that is not None and other is None:
        return that

    elif that is None and other is not None:
        return other

    elif that is None and other is None:
        return None

    else:
        raise AssertionError("Unhandled execution path")


def _merge_constraints(
    that: Optional[Constraints], other: Optional[Constraints]
) -> Optional[Constraints]:
    """Combine the constraints on tighter bounds."""
    if that is not None and other is not None:
        return Constraints(
            len_constraint=_merge_len_constraints(
                that.len_constraint, other.len_constraint
            ),
            patterns=_merge_pattern_constraints(that.patterns, other.patterns),
            set_of_primitives=_merge_set_of_primitives_constraints(
                that.set_of_primitives, other.set_of_primitives
            ),
            set_of_enumeration_literals=_merge_set_of_enumeration_literals_constraints(
                that.set_of_enumeration_literals, other.set_of_enumeration_literals
            ),
        )

    elif that is not None and other is None:
        return that

    elif that is None and other is not None:
        return other

    elif that is None and other is None:
        return None

    else:
        raise AssertionError("Unhandled execution path")


def _infer_constraints_of_constrained_primitive_without_inheritance(
    constrained_primitive: intermediate.ConstrainedPrimitive,
    pattern_verifications_by_name: infer_for_schema_pattern.PatternVerificationsByName,
) -> Tuple[Optional[Constraints], Optional[List[Error]]]:
    """
    Infer the constraints from the invariants on self of the constrained primitive.

    We do not go up (or down) the inheritance tree -- the constraints from the parents
    are not inherited at this step.

    If there are no constraints inferred from the constrained primitive, we return None.
    """
    errors = []  # type: List[Error]

    len_constraint: Optional[LenConstraint] = None

    if constrained_primitive.constrainee in infer_for_schema_len.LENGTHABLE_PRIMITIVES:
        (
            len_constraint,
            len_constraint_errors,
        ) = infer_for_schema_len.infer_len_constraint_of_self(
            constrained_primitive=constrained_primitive
        )

        if len_constraint_errors is not None:
            errors.extend(len_constraint_errors)
        else:
            assert len_constraint is not None

            # NOTE (mristin):
            # We do not want to keep dummy constraints.
            if len_constraint.min_value is None and len_constraint.max_value is None:
                len_constraint = None

    pattern_constraints: Optional[Sequence[PatternConstraint]] = None

    if constrained_primitive.constrainee is intermediate.PrimitiveType.STR:
        pattern_constraints = infer_for_schema_pattern.infer_patterns_on_self(
            constrained_primitive=constrained_primitive,
            pattern_verifications_by_name=pattern_verifications_by_name,
        )

        if len(pattern_constraints) == 0:
            # NOTE (mristin):
            # We do not want to keep dummy constraints.
            pattern_constraints = None

    if len(errors) > 0:
        return None, errors

    if len_constraint is None and pattern_constraints is None:
        return None, None

    return (
        Constraints(
            len_constraint=len_constraint,
            patterns=pattern_constraints,
            # NOTE (mristin):
            # We do not match the set of primitives on the constrained primitives at the
            # moment, but this could be easily implemented.
            set_of_primitives=None,
            set_of_enumeration_literals=None,
        ),
        None,
    )


@ensure(lambda result: not (result[1] is not None) or len(result[1]) > 0)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _infer_constraints_by_constrained_primitive(
    symbol_table: intermediate.SymbolTable,
    pattern_verifications_by_name: infer_for_schema_pattern.PatternVerificationsByName,
) -> Tuple[
    Optional[MutableMapping[intermediate.ConstrainedPrimitive, Constraints]],
    Optional[List[Error]],
]:
    """
    Infer the constraints from the constrained primitives considering the inheritance.

    If there are no constraints for the given constrained primitive, it is not present
    in the mapping.
    """
    errors: List[Error] = []

    mapping: MutableMapping[intermediate.ConstrainedPrimitive, Constraints] = dict()

    # NOTE (mristin):
    # We perform the first pass where we disregard inheritance and return if there are
    # any errors. We can then be certain that there will be no errors when we stack
    # the constraints considering the inheritance tree.

    for constrained_primitive in symbol_table.constrained_primitives:
        (
            constraints,
            errors_constrained_primitive,
        ) = _infer_constraints_of_constrained_primitive_without_inheritance(
            constrained_primitive=constrained_primitive,
            pattern_verifications_by_name=pattern_verifications_by_name,
        )

        if errors_constrained_primitive is not None:
            errors.append(
                Error(
                    constrained_primitive.parsed.node,
                    f"Failed to infer the schema constraints "
                    f"from constrained primitive {constrained_primitive.name!r}",
                    errors_constrained_primitive,
                )
            )

        if constraints is not None:
            mapping[constrained_primitive] = constraints

    if len(errors) > 0:
        return None, errors

    for our_type in symbol_table.our_types_topologically_sorted:
        if not isinstance(our_type, intermediate.ConstrainedPrimitive):
            continue

        # NOTE (mristin):
        # We rename for clarity when the reader reads the previous code.
        constrained_primitive = our_type

        constraints = mapping.get(constrained_primitive, None)

        # NOTE (mristin):
        # The topological order ensures that we have processed the parents already.
        for parent in constrained_primitive.inheritances:
            # NOTE (mristin):
            # The constrained primitive inherits all the constraints from the parent,
            # and then it might tighten them some more.
            constraints = _merge_constraints(mapping.get(parent, None), constraints)

        if constraints is not None:
            mapping[constrained_primitive] = constraints

    return mapping, None


def _over_non_optional_type_annotations(
    type_annotation: intermediate.TypeAnnotationUnion,
) -> Iterator[intermediate.TypeAnnotationExceptOptional]:
    """
    Iterate recursively over the type annotation and all its nested type annotations.

    The optional type annotations are recursed into, but will not be yielded.
    """
    if not isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        yield type_annotation

    if isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        yield from _over_non_optional_type_annotations(type_annotation.value)

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        yield from _over_non_optional_type_annotations(type_annotation.items)

    elif isinstance(
        type_annotation,
        (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
    ):
        pass

    else:
        # noinspection PyTypeChecker
        assert_never(type_annotation)


@ensure(
    lambda result: not (result[0] is not None)
    or (all(not constraints.is_empty() for constraints in result[0].values()))
)
@ensure(lambda result: not (result[1] is not None) or len(result[1]) >= 1)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _infer_constraints_of_class_values_without_inheritance(
    cls: intermediate.ClassUnion,
    constraints_by_constrained_primitive: Mapping[
        intermediate.ConstrainedPrimitive, Constraints
    ],
    pattern_verifications_by_name: infer_for_schema_pattern.PatternVerificationsByName,
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[MutableConstraintsByValue], Optional[List[Error]],]:
    """
    Infer the constraints on all the possible values of the class.

    The class hierarchy is not taken into account; we consider only the invariants
    defined for this class.
    """
    errors: List[Error] = []

    mapping: MutableConstraintsByValue = dict()

    # region Constraints on length
    (
        len_constraints_from_invariants,
        some_errors,
    ) = infer_for_schema_len.len_constraints_from_invariants(cls=cls)

    if some_errors is not None:
        errors.extend(some_errors)
    else:
        assert len_constraints_from_invariants is not None

        for prop, len_constraint in len_constraints_from_invariants.items():
            if len_constraint.min_value is None and len_constraint.max_value is None:
                continue

            type_anno = intermediate.beneath_optional(prop.type_annotation)

            merged_constraints = _merge_constraints(
                mapping.get(type_anno, None), Constraints(len_constraint=len_constraint)
            )

            if merged_constraints is not None:
                mapping[type_anno] = merged_constraints

    # endregion

    # region Pattern constraints

    patterns_from_invariants_by_property = (
        infer_for_schema_pattern.patterns_from_invariants(
            cls=cls,
            pattern_verifications_by_name=pattern_verifications_by_name,
        )
    )

    for prop, pattern_constraints in patterns_from_invariants_by_property.items():
        if len(pattern_constraints) == 0:
            continue

        type_anno = intermediate.beneath_optional(prop.type_annotation)

        merged_constraints = _merge_constraints(
            mapping.get(type_anno, None), Constraints(patterns=pattern_constraints)
        )

        if merged_constraints is not None:
            mapping[type_anno] = merged_constraints

    # endregion

    # region Constraints from constant sets

    # fmt: off
    set_constraints, some_errors = (
        infer_for_schema_set.infer_set_constraints_by_property_from_invariants(
            cls=cls,
            symbol_table=symbol_table
        )
    )
    # fmt: on

    if some_errors is not None:
        errors.extend(some_errors)
    else:
        assert set_constraints is not None

        for (
            prop,
            set_of_primitives_constraint,
        ) in set_constraints.set_of_primitives_by_property.items():
            type_anno = intermediate.beneath_optional(prop.type_annotation)

            merged_constraints = _merge_constraints(
                mapping.get(type_anno, None),
                Constraints(set_of_primitives=set_of_primitives_constraint),
            )

            if merged_constraints is not None:
                mapping[type_anno] = merged_constraints

        for (
            prop,
            set_of_enumeration_literals_constraint,
        ) in set_constraints.set_of_enumeration_literals_by_property.items():
            type_anno = intermediate.beneath_optional(prop.type_annotation)

            merged_constraints = _merge_constraints(
                mapping.get(type_anno, None),
                Constraints(
                    set_of_enumeration_literals=set_of_enumeration_literals_constraint
                ),
            )

            if merged_constraints is not None:
                mapping[type_anno] = merged_constraints

    # endregion

    if len(errors) > 0:
        return None, errors

    # region Patterns from constrained primitives

    for prop in cls.properties:
        for type_anno in _over_non_optional_type_annotations(prop.type_annotation):
            if not isinstance(type_anno, intermediate.OurTypeAnnotation):
                continue

            if not isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                continue

            constraints = _merge_constraints(
                mapping.get(type_anno, None),
                constraints_by_constrained_primitive.get(type_anno.our_type, None),
            )

            if constraints is not None and not constraints.is_empty():
                mapping[type_anno] = constraints

    # endregion

    assert len(errors) == 0
    return mapping, None


@ensure(lambda result: not (result[1] is not None) or len(result[1]) > 0)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def infer_constraints_by_class(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[
    Optional[Mapping[intermediate.ClassUnion, ConstraintsByValue]],
    Optional[List[Error]],
]:
    """Infer the constraints from the invariants and constrained primitives."""
    errors = []  # type: List[Error]

    pattern_verifications_by_name = (
        infer_for_schema_pattern.map_pattern_verifications_by_name(
            verifications=symbol_table.verification_functions
        )
    )

    (
        constraints_by_constrained_primitive,
        some_errors,
    ) = _infer_constraints_by_constrained_primitive(
        symbol_table=symbol_table,
        pattern_verifications_by_name=pattern_verifications_by_name,
    )

    if some_errors is not None:
        errors.extend(some_errors)

    if len(errors) > 0:
        return None, errors

    assert constraints_by_constrained_primitive is not None

    mapping: MutableMapping[intermediate.ClassUnion, MutableConstraintsByValue] = dict()

    for cls in symbol_table.classes:
        (
            constraints_by_value,
            some_errors,
        ) = _infer_constraints_of_class_values_without_inheritance(
            cls=cls,
            constraints_by_constrained_primitive=constraints_by_constrained_primitive,
            pattern_verifications_by_name=pattern_verifications_by_name,
            symbol_table=symbol_table,
        )

        if some_errors is not None:
            errors.extend(some_errors)
            continue

        assert constraints_by_value is not None

        mapping[cls] = constraints_by_value

    if len(errors) > 0:
        return None, errors

    # NOTE (mristin):
    # We stack now the constraints considering the class hierarchy. The topological
    # order guarantees that the parents have been processed before the children.
    for our_type in symbol_table.our_types_topologically_sorted:
        if not isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            continue

        # NOTE (mristin):
        # We introduce the alias for better readability with the previous section.
        cls = our_type

        that_constraints_by_value = mapping[cls]

        for parent in cls.inheritances:
            parent_constraints_by_value = mapping[parent]

            # NOTE (mristin):
            # The class inherits all the constraints from the parent, and then it might
            # tighten them some more.
            for type_anno, parent_constraints in parent_constraints_by_value.items():
                merged_constraints = _merge_constraints(
                    parent_constraints,
                    that_constraints_by_value.get(type_anno, None),
                )

                if merged_constraints is not None:
                    that_constraints_by_value[type_anno] = merged_constraints

    return mapping, None


def tightening_steps_from_other_to_that_constraints(
    that: Constraints, other: Optional[Constraints]
) -> Optional[Constraints]:
    """
    Identify constraints to be updated to go from ``other`` to ``that``.

    This function is used to spot tightening of the constraints in the children
    classes which override the parents' constraints.

    We assume that the in-lining and merging of the constraints has been already
    performed before. Any constraints in the ``other`` which differs from ``that``
    constraint is assumed to be "overridden" (*i.e.*, tightened) by ``that`` constraint.

    You can think of the resulting constraint list as all the constraints that a child
    class (corresponding to ``that``) imposes in addition to constraints that are already
    imposed by the parent class (corresponding to ``other``).
    """
    if other is None:
        return that

    len_constraint: Optional[LenConstraint] = None

    if other.len_constraint is not None:
        assert that.len_constraint is not None, (
            "The ``other``, corresponding to a parent class, defines a length constraint, "
            "but the child class, corresponding to ``that``, relaxes the constraint and "
            "defines no length constraints. This breaks the behavioral subtyping. "
            "Have the constraints been properly merged and inlined?\n"
            f"{other.len_constraint=}"
        )

        len_constraint = (
            that.len_constraint
            if not that.len_constraint.equals(other.len_constraint)
            else None
        )
    else:
        len_constraint = that.len_constraint

    patterns: Optional[Sequence[PatternConstraint]] = None

    if other.patterns is not None:
        assert that.patterns is not None, (
            "The ``other``, corresponding to a parent class, defines patterns, "
            "but the child class, corresponding to ``that``, relaxes the constraint and "
            "defines no patterns. This breaks the behavioral subtyping. "
            "Have the constraints been properly merged and inlined?\n"
            f"{other.patterns=}, {that.patterns=}"
        )

        other_pattern_set = set(
            pattern_constraint.pattern for pattern_constraint in other.patterns
        )

        that_pattern_set = set(
            pattern_constraint.pattern for pattern_constraint in that.patterns
        )

        assert other_pattern_set.intersection(that_pattern_set) == other_pattern_set, (
            "The ``other``, corresponding to a parent class, defines more patterns "
            "than the child class, corresponding to ``that`` -- this amounts to relaxing "
            "constraints. This breaks the behavioral subtyping. "
            "Have the constraints been properly merged and inlined?\n"
            f"{other.patterns=}, {that.patterns=}"
        )

        # NOTE (mristin):
        # We select only patterns which are not already defined in ``other``, as
        # the ``other`` corresponds to the parent class.
        patterns = [
            pattern_constraint
            for pattern_constraint in that.patterns
            if pattern_constraint.pattern not in other_pattern_set
        ]

        if len(patterns) == 0:
            patterns = None
    else:
        patterns = that.patterns

    set_of_primitives: Optional[SetOfPrimitivesConstraint] = None

    if other.set_of_primitives is not None:
        assert that.set_of_primitives is not None, (
            "The ``other``, corresponding to a parent class, defines set of primitives, "
            "but the child class, corresponding to ``that``, relaxes the constraint and "
            "defines no set of primitives. This breaks the behavioral subtyping. "
            "Have the constraints been properly merged and inlined?\n"
            f"{other.set_of_primitives=}"
        )

        other_set_of_primitives = set(
            literal.value for literal in other.set_of_primitives.literals
        )

        that_set_of_primitives = set(
            literal.value for literal in that.set_of_primitives.literals
        )

        assert (
            that_set_of_primitives.intersection(other_set_of_primitives)
            == that_set_of_primitives
        ), (
            "The ``that``, corresponding to a child class, can only tighten the set of "
            "allowed primitives from ``other``, which corresponds to the parent class. "
            "However, ``that`` defines more literals -- this breaks the behavioral "
            "subtyping. Have the constraints been properly merged and inlined?\n"
            f"{other.set_of_primitives=}, {that.set_of_primitives=}"
        )

        if not that.set_of_primitives.equals(other.set_of_primitives):
            set_of_primitives = that.set_of_primitives

    else:
        set_of_primitives = that.set_of_primitives

    set_of_enumeration_literals: Optional[SetOfEnumerationLiteralsConstraint] = None

    if other.set_of_enumeration_literals is not None:
        assert that.set_of_enumeration_literals is not None, (
            "The ``other``, corresponding to a parent class, defines set of "
            "enumeration literals, but the child class, corresponding to ``that``, "
            "relaxes the constraint and defines no set of enumeration literals. This "
            "breaks the behavioral subtyping. Have the constraints been properly merged "
            "and inlined?\n"
            f"{other.set_of_enumeration_literals=}"
        )

        other_set_of_enumeration_literals = set(
            literal.value for literal in other.set_of_enumeration_literals.literals
        )

        that_set_of_enumeration_literals = set(
            literal.value for literal in that.set_of_enumeration_literals.literals
        )

        assert (
            that_set_of_enumeration_literals.intersection(
                other_set_of_enumeration_literals
            )
            == that_set_of_enumeration_literals
        ), (
            "The ``that``, corresponding to a child class, can only tighten the set of "
            "allowed enumeration literals from ``other``, which corresponds to "
            "the parent class. However, ``that`` defines more literals -- this breaks "
            "the behavioral subtyping. Have the constraints been properly merged and "
            "inlined?\n"
            f"{other.set_of_enumeration_literals=}, {that.set_of_enumeration_literals=}"
        )

        if not that.set_of_enumeration_literals.equals(
            other.set_of_enumeration_literals
        ):
            set_of_enumeration_literals = that.set_of_enumeration_literals
    else:
        set_of_enumeration_literals = that.set_of_enumeration_literals

    tightening = Constraints(
        len_constraint=len_constraint,
        patterns=patterns,
        set_of_primitives=set_of_primitives,
        set_of_enumeration_literals=set_of_enumeration_literals,
    )

    return tightening



"""Generate code for common JSON de/serialization shared across the tests."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.golang import common as golang_common
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(repo_url: Stripped) -> str:
    """Generate code for common JSON de/serialization shared across the tests."""
    blocks = [
        Stripped(
            """\
package jsonization_test"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"encoding/json"
{I}"fmt"
{I}"os"
{I}"path/filepath"
{I}"strings"
{I}"testing"
{I}aasjsonization "{repo_url}/jsonization"
{I}aastesting "{repo_url}/aastesting"
)"""
        ),
        Stripped(
            f"""\
// Assert that there is no de-serialization error when de-serializing
// from `source`.
func assertNoDeserializationError(
{I}t *testing.T,
{I}err error,
{I}source string,
) (ok bool) {{
{I}ok = true
{I}if err != nil {{
{II}ok = false
{II}t.Fatalf(
{III}"Expected no de-serialization error from %s, "+
{IIII}"but got: %v",
{III}source, err,
{II})
{II}return
{I}}}
{I}return
}}"""
        ),
        Stripped(
            f"""\
// Assert that there is no serialization error when serializing the instance
// originally coming from `source`.
func assertNoSerializationError(
{I}t *testing.T,
{I}err error,
{I}source string,
) (ok bool) {{
{I}ok = true
{I}if err != nil {{
{II}ok = false
{II}t.Fatalf(
{III}"Expected no serialization error when serializing "+
{IIII}"the instance obtained from %s, "+
{IIII}"but got: %v",
{III}source, err,
{II})
{II}return
{I}}}
{I}return
}}"""
        ),
        Stripped(
            f"""\
// Assert that the serialization `other`, as JSON-able, equals the original
// JSON-able `that` read from `source`.
func assertSerializationEqualsDeserialization(
{I}t *testing.T,
{I}that interface{{}},
{I}other interface{{}},
{I}source string,
) (ok bool) {{
{I}ok = true

{I}thatBytes, err := json.Marshal(that)
{I}if err != nil {{
{II}panic(fmt.Sprintf("Failed to marshal that jsonable %v: %s", that, err.Error()))
{I}}}
{I}thatText := string(thatBytes)

{I}otherBytes, err := json.Marshal(other)
{I}if err != nil {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Failed to marshal other jsonable %v: %s", other, err.Error(),
{III}),
{II})
{I}}}
{I}otherText := string(otherBytes)

{I}if thatText != otherText {{
{II}ok = false
{II}t.Fatalf(
{III}"The serialization of the de-serialized instance from %s does not equal "+
{IIII}"the original JSON-able:\\nOriginal:\\n%s\\nSerialized:\\n%s",
{III}source, thatText, otherText,
{II})
{II}return
{I}}}

{I}return
}}"""
        ),
        Stripped(
            f"""\
// Assert that there is a de-serialization error.
//
// If [aastesting.RecordMode] is set, the de-serialization error is re-recorded
// to `expectedPth`. Otherwise, the error is compared against the golden file
// `expectedPth`.
func assertDeserializationErrorEqualsExpectedOrRecord(
{I}t *testing.T,
{I}err error,
{I}source string,
{I}expectedPth string,
) (ok bool) {{
{I}ok = true

{I}if err == nil {{
{II}ok = false
{II}t.Fatalf("De-serialization error expected from %s, but got none", source)
{II}return
{I}}}

{I}var deseriaErr *aasjsonization.DeserializationError
{I}deseriaErr, ok = err.(*aasjsonization.DeserializationError)
{I}if !ok {{
{II}t.Fatalf("Expected a de-serialization error, but got: %v from %s", err, source)
{II}return
{I}}}

{I}// Add a new line for POSIX systems.
{I}got := deseriaErr.Error() + "\\n"

{I}if aastesting.RecordMode {{
{II}parent := filepath.Dir(expectedPth)
{II}err := os.MkdirAll(parent, os.ModePerm)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to create the directory %s: %s", parent, err.Error(),
{IIII}),
{III})
{II}}}

{II}err = os.WriteFile(expectedPth, []byte(got), 0644)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to write to the file %s: %s", expectedPth, err.Error(),
{IIII}),
{III})
{II}}}
{I}}} else {{
{II}_, err := os.Stat(expectedPth)
{II}if err != nil {{
{III}ok = false
{III}t.Fatalf(
{IIII}"Failed to stat the file %s: %s; if the file does not exist, "+
{IIIII}"you probably want to record the test data by "+
{IIIII}"setting the environment variable %s",
{IIII}expectedPth, err.Error(), aastesting.RecordModeEnvironmentVariableName,
{III})
{III}return
{II}}}

{II}bb, err := os.ReadFile(expectedPth)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to read from file %s: %s", expectedPth, err.Error(),
{IIII}),
{III})
{II}}}

{II}expected := string(bb)

{II}// NOTE (mristin):
{II}// Git automatically strips and adds `\\r`, so we have to remove it here
{II}// to obtain a canonical text.
{II}expected = strings.Replace(expected, "\\r", "", -1)

{II}if expected != got {{
{III}ok = false
{III}t.Fatalf(
{IIII}"What we got differs from the expected in %s. "+
{IIIII}"We got:\\n%s\\nWe expected:\\n%s",
{IIII}expectedPth, got, expected,
{III})
{III}return
{II}}}
{I}}}

{I}return
}}"""
        ),
        golang_common.WARNING,
    ]  # type: List[Stripped]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the unit test code for Golang."""

from aas_core_codegen.golang.tests import (
    _generate_aastesting_common_jsonization,
    _generate_aastesting_constants,
    _generate_aastesting_deep_equal,
    _generate_aastesting_doc,
    _generate_aastesting_filesystem,
    _generate_aastesting_tracing,
    _generate_descend_test_common,
    _generate_descend_test_descend_once_test,
    _generate_descend_test_descend_test,
    _generate_enhancing_test,
    _generate_is_xxx_test,
    _generate_jsonization_test_classes_with_descendants_test,
    _generate_jsonization_test_common_test,
    _generate_jsonization_test_concrete_classes_test,
    _generate_jsonization_test_enums_test,
    _generate_verification_test,
    _generate_xmlization_test_concrete_classes_test,
    _generate_xmlization_test_common_test,
    _generate_xxx_or_default_test,
)

generate_aastesting_common_jsonization = (
    _generate_aastesting_common_jsonization.generate
)
generate_aastesting_constants = _generate_aastesting_constants.generate
generate_aastesting_deep_equal = _generate_aastesting_deep_equal.generate
generate_aastesting_doc = _generate_aastesting_doc.generate
generate_aastesting_filesystem = _generate_aastesting_filesystem.generate
generate_aastesting_tracing = _generate_aastesting_tracing.generate
generate_descend_test_common = _generate_descend_test_common.generate
generate_descend_test_descend_once_test = (
    _generate_descend_test_descend_once_test.generate
)
generate_descend_test_descend_test = _generate_descend_test_descend_test.generate
generate_enhancing_test = _generate_enhancing_test.generate
generate_is_xxx_test = _generate_is_xxx_test.generate
generate_jsonization_test_classes_with_descendants_test = (
    _generate_jsonization_test_classes_with_descendants_test.generate
)
generate_jsonization_test_common_test = _generate_jsonization_test_common_test.generate
generate_jsonization_test_concrete_classes_test = (
    _generate_jsonization_test_concrete_classes_test.generate
)
generate_jsonization_test_enums_test = _generate_jsonization_test_enums_test.generate
generate_verification_test = _generate_verification_test.generate
generate_xmlization_test_concrete_classes_test = (
    _generate_xmlization_test_concrete_classes_test.generate
)
generate_xmlization_test_common_test = _generate_xmlization_test_common_test.generate
generate_xxx_or_default_test = _generate_xxx_or_default_test.generate



"""Generate code to test enhancing the model instances."""

from typing import List

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test enhancing the model instances.
    """
    blocks = [
        Stripped(
            f"""\
public static class Enhancement {{
{I}public final long someCustomId;

{I}public Enhancement(long someCustomId) {{
{II}this.someCustomId = someCustomId;
{I}}}
}}"""
        ),
        Stripped(
            f"""\
private Enhancer<Enhancement> createEnhancer() {{
{I}AtomicLong lastCustomId = new AtomicLong();

{I}Function<IClass, Optional<Enhancement>> enhancementFactory =
{II}iClass -> Optional.of(new Enhancement(lastCustomId.incrementAndGet()));

{I}return new Enhancer<>(enhancementFactory);
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name = java_naming.class_name(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name}() throws IOException {{
{I}final {cls_name} instance = CommonJsonization.loadMaximal{cls_name}();

{I}final Enhancer<Enhancement> enhancer = createEnhancer();

{I}assert !enhancer.unwrap(instance).isPresent();

{I}final IClass wrapped = enhancer.wrap(instance);
{I}assertNotNull(wrapped);

{I}final Set<Long> idSet = new HashSet<>();
{I}idSet.add(enhancer.mustUnwrap(wrapped).someCustomId);
{I}wrapped
{II}.descend()
{II}.forEach(descendant -> idSet.add(enhancer.mustUnwrap(descendant).someCustomId));

{I}assertFalse(enhancer.unwrap(instance).isPresent());
{I}assertNotNull(wrapped);
{I}assertEquals(
{II}1L,
{II}idSet.stream()
{III}.min(Comparator.comparing(Long::valueOf))
{III}.orElseThrow(() -> new IllegalStateException("Missing min value for wrapped.")));
{I}assertEquals(
{II}idSet.size(),
{II}idSet.stream()
{III}.max(Comparator.comparing(Long::valueOf))
{III}.orElseThrow(() -> new IllegalStateException("Missing max value for wrapped.")));
}} // public void test{cls_name}"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestEnhancing.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.*;

import {package}.enhancing.Enhancer;
import {package}.types.impl.*;
import {package}.types.model.IClass;
import java.io.IOException;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Function;
import org.junit.jupiter.api.Test;

public class TestEnhancing {{
{I}{indent_but_first_line(blocks_joined, I)}
}}  // class TestEnhancing
// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test ``*OrDefault`` functions."""


import io
from typing import List, Optional

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate code to test ``*OrDefault`` functions."""
    blocks = [
        Stripped("package types_xxx_or_default_test"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"path/filepath"
{I}"fmt"
{I}"encoding/json"
{I}"os"
{I}"reflect"
{I}"strings"
{I}"testing"
{I}aasstringification "{repo_url}/stringification"
{I}aastesting "{repo_url}/aastesting"
{I}aastypes "{repo_url}/types"
)"""
        ),
        Stripped(
            f"""\
// Represent explicitly a literal of an enumeration.
type enumerationLiteral struct {{
{I}enumerationName string
{I}literalName string
}}"""
        ),
        Stripped(
            f"""\
func (el *enumerationLiteral) String() string {{
{I}return fmt.Sprintf("%s.%s", el.enumerationName, el.literalName)
}}"""
        ),
        Stripped(
            f"""\
// Marshal the value as JSON, or panic otherwise.
func mustJSONMarshal(value interface{{}}) string {{
{I}bb, err := json.Marshal(value)
{I}if err != nil {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Failed to marshal the value %v to JSON: %s",
{IIII}value, err.Error(),
{III}),
{II})
{I}}}

{I}return string(bb)
}}"""
        ),
        Stripped(
            f"""\
func stringify(value interface{{}}) (got string) {{
{I}if value == nil {{
{II}got = mustJSONMarshal(value)
{I}}} else {{
{II}// See: https://stackoverflow.com/questions/38748098/golang-type-switch-how-to-match-a-generic-slice-array-map-chan
{II}reflected := reflect.ValueOf(value)

{II}if reflected.Kind() == reflect.Slice {{
{III}parts := make([]string, reflected.Len())

{III}for i := 0; i < reflected.Len(); i++ {{
{IIII}item := reflected.Index(i)
{IIII}parts[i] = stringify(item)
{III}}}

{III}got = fmt.Sprintf("[%s]", strings.Join(parts, ", "))
{II}}} else {{
{III}switch casted := value.(type) {{
{III}case bool:
{IIII}got = mustJSONMarshal(casted)
{III}case int:
{IIII}got = mustJSONMarshal(casted)
{III}case string:
{IIII}got = mustJSONMarshal(casted)
{III}case []byte:
{IIII}got = fmt.Sprintf("%d byte(s)", len(casted))
{III}case *enumerationLiteral:
{IIII}got = casted.String()
{III}case aastypes.IClass:
{IIII}got = aastesting.TraceMark(casted)
{III}default:
{IIII}panic(
{IIII}{I}fmt.Sprintf(
{IIII}{II}"We do not know hot to represent the value of type %T: %v",
{IIII}{II}value, value,
{IIII}{I}),
{IIII})
{III}}}
{II}}}
{I}}}

{I}return
}}"""
        ),
        Stripped(
            f"""\
// Represent `value` such that we can immediately check whether it is the default value
// or the set one.
//
// We compare it against the recorded golden file, if not [aastesting.RecordMode].
// If there are differences, a `message` is set.
//
// Otherwise, when [aastesting.RecordMode] is set, we re-record the golden file.
func compareOrRerecordValue(
{I}value interface{{}},
{I}expectedPath string,
) (message *string) {{
{I}got := stringify(value)

{I}// NOTE (mristin):
{I}// Add a new line for POSIX systems.
{I}got += "\\n"

{I}if aastesting.RecordMode {{
{II}parent := filepath.Dir(expectedPath)
{II}err := os.MkdirAll(parent, os.ModePerm)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIII}{I}"Failed to create the directory %s: %s", parent, err.Error(),
{IIII}),
{III})
{II}}}

{II}err = os.WriteFile(expectedPath, []byte(got), 0644)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIII}{I}"Failed to write to the file %s: %s", expectedPath, err.Error(),
{IIII}),
{III})
{II}}}
{I}}} else {{
{II}bb, err := os.ReadFile(expectedPath)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIII}{I}"Failed to read from file %s: %s", expectedPath, err.Error(),
{IIII}),
{III})
{II}}}

{II}expected := string(bb)

{II}// NOTE (mristin):
{II}// Git automatically strips and adds `\\r`, so we have to remove it here
{II}// to obtain a canonical text.
{II}expected = strings.Replace(expected, "\\r", "", -1)

{II}if expected != got {{
{III}text := fmt.Sprintf(
{IIII}"What we got differs from the expected in %s. " +
{IIII}"We got:\\n%s\\nWe expected:\\n%s",
{IIII}expectedPath, got, expected,
{III})
{III}message = &text
{II}}}
{I}}}

{I}return
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        x_or_default_methods = [
            method
            for method in concrete_cls.methods
            if method.name.endswith("_or_default")
        ]  # type: List[intermediate.MethodUnion]

        model_type = naming.json_model_type(concrete_cls.name)

        for method in x_or_default_methods:
            method_name = golang_naming.method_name(method.name)

            result_enum = None  # type: Optional[intermediate.Enumeration]
            assert method.returns is not None, (
                f"Expected all X_or_default to return something, "
                f"but got None for {concrete_cls}.{method.name}"
            )

            if isinstance(
                method.returns, intermediate.OurTypeAnnotation
            ) and isinstance(method.returns.our_type, intermediate.Enumeration):
                result_enum = method.returns.our_type

            if result_enum is None:
                value_assignment_snippet = Stripped(
                    f"value := instance.{method_name}()"
                )
            else:
                enum_to_string_name = golang_naming.function_name(
                    Identifier(f"must_{result_enum.name}_to_string")
                )

                enum_name = golang_naming.enum_name(result_enum.name)

                value_assignment_snippet = Stripped(
                    f"""\
value := &enumerationLiteral{{
{I}enumerationName: {golang_common.string_literal(enum_name)},
{I}literalName: aasstringification.{enum_to_string_name}(
{II}instance.{method_name}(),
{I}),
}}"""
                )

            test_function_name = golang_naming.function_name(
                Identifier(f"Test_{concrete_cls.name}_{method.name}_on_minimal")
            )

            must_load_minimal_name = golang_naming.function_name(
                Identifier(f"must_load_minimal_{concrete_cls.name}")
            )

            blocks.append(
                Stripped(
                    f"""\
func {test_function_name}(t *testing.T) {{
{I}instance := aastesting.{must_load_minimal_name}()

{I}{indent_but_first_line(value_assignment_snippet, I)}

{I}expectedPth := filepath.Join(
{II}aastesting.TestDataDir,
{II}"XxxOrDefault",
{II}{golang_common.string_literal(model_type)},
{II}"{method_name}.on_minimal.txt",
{I})

{I}message := compareOrRerecordValue(
{II}value,
{II}expectedPth,
{I})

{I}if message != nil {{
{II}t.Fatal(*message)
{I}}}
}}"""
                )
            )

            test_function_name = golang_naming.function_name(
                Identifier(f"Test_{concrete_cls.name}_{method.name}_on_maximal")
            )

            must_load_maximal_name = golang_naming.function_name(
                Identifier(f"must_load_maximal_{concrete_cls.name}")
            )

            blocks.append(
                Stripped(
                    f"""\
func {test_function_name}(t *testing.T) {{
{I}instance := aastesting.{must_load_maximal_name}()

{I}{indent_but_first_line(value_assignment_snippet, I)}

{I}expectedPth := filepath.Join(
{II}aastesting.TestDataDir,
{II}"XxxOrDefault",
{II}{golang_common.string_literal(model_type)},
{II}"{method_name}.on_maximal.txt",
{I})

{I}message := compareOrRerecordValue(
{II}value,
{II}expectedPth,
{I})

{I}if message != nil {{
{II}t.Fatal(*message)
{I}}}
}}"""
                )
            )

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the JSON de/serialization of concrete classes."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the JSON de/serialization of concrete classes."""
    blocks = [
        Stripped(
            """\
/**
 * Test JSON de/serialization of concrete classes.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as fs from "fs";
import * as path from "path";

import * as AasJsonization from "../src/jsonization";
import * as AasTypes from "../src/types";
import * as AasVerification from "../src/verification";

import * as TestCommon from "./common";"""
        ),
        Stripped(
            f"""\
/**
 * Assert that the result of the chain JSON → de-serialize → object → serialize → JSON
 * gives the input.
 */
function assertSerializeDeserializeEqualsOriginal(
{I}originalJsonable: AasJsonization.JsonValue,
{I}instance: AasTypes.Class,
{I}aPath: string
): void {{
{I}let jsonable: AasJsonization.JsonValue | null = null;
{I}try {{
{II}jsonable = AasJsonization.toJsonable(instance);
{I}}} catch (error) {{
{II}throw new Error(
{III}"Expected no exception during JSON serialization " +
{III}`of an instance of ${{instance.constructor.name}} from ${{aPath}}, ` +
{III}`but got: ${{error}}`
{II});
{I}}}

{I}const inequalityError = TestCommon.checkJsonablesEqual(
{II}originalJsonable,
{II}jsonable
{I});
{I}if (inequalityError !== null) {{
{II}throw new Error(
{III}`The original JSON from ${{aPath}} is unequal the serialized JSON: ` +
{III}`${{inequalityError.path}}: ${{inequalityError.message}}`
{II});
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Assert that the deserialization error equals the expected golden one,
 * or, if {{@link common.RECORD_MODE}} set, re-record the expected error.
 *
 * @param error - obtained error during the de-serialization
 * @param aPath - to the JSON file which caused the de-serialization error
 * @throws an {{@link Error}} if assertion fails
 */
function assertDeserializationErrorEqualsExpectedOrRecord(
{I}error: AasJsonization.DeserializationError,
{I}aPath: string
): void {{
{I}const errorPath = aPath + ".error";
{I}const got = `${{error.path}}: ${{error.message}}\\n`;

{I}if (TestCommon.RECORD_MODE) {{
{II}fs.writeFileSync(errorPath, got, "utf-8");
{I}}} else {{
{II}if (!fs.existsSync(errorPath)) {{
{III}throw new Error(
{IIII}`The file with the recorded deserialization error does ` +
{IIII}`not exist: ${{errorPath}}; you probably want to set ` +
{IIII}`the environment variable ${{TestCommon.RECORD_MODE_ENVIRONMENT_VARIABLE_NAME}}?`
{III});
{II}}}

{II}const expected =
{III}fs.readFileSync(errorPath, "utf-8")
{III}.replace(/\\r\\n/g, "\\n");
{II}if (expected !== got) {{
{III}throw new Error(
{IIII}`Expected the error:\\n${{JSON.stringify(expected)}}\\n, ` +
{IIII}`but got:\\n${{JSON.stringify(got)}}\\n` +
{IIII}`when de-serializing from ${{aPath}}`
{III});
{II}}}
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Assert that the obtained verification errors equal the expected verification errors,
 * or, if {{@link common.RECORD_MODE}} set, re-record the expected errors.
 *
 * @param errors - obtained verification errors
 * @param aPath - to the JSON file which caused the verification errors
 * @throws an {{@link Error}} if assertion fails
 */
function assertVerificationErrorsEqualExpectedOrRecord(
{I}errors: Array<AasVerification.VerificationError>,
{I}aPath: string
): void {{
{I}const errorsPath = aPath + ".errors";

{I}const lines = new Array<string>();
{I}for (const error of errors) {{
{II}lines.push(`${{error.path}}: ${{error.message}}`);
{I}}}
{I}// NOTE (mristin):
{I}// We add a new line for POSIX systems which prefer a new line
{I}// at the end of the file.
{I}lines.push("");
{I}const got = lines.join("\\n");

{I}if (TestCommon.RECORD_MODE) {{
{II}fs.writeFileSync(errorsPath, got, "utf-8");
{I}}} else {{
{II}if (!fs.existsSync(errorsPath)) {{
{III}throw new Error(
{IIII}`The file with the recorded verification errors ` +
{IIII}`does not exist: ${{errorsPath}}; you probably want to set the environment ` +
{IIII}`variable ${{TestCommon.RECORD_MODE_ENVIRONMENT_VARIABLE_NAME}}?`
{III});
{II}}}

{II}const expected =
{III}fs.readFileSync(errorsPath, "utf-8")
{III}.replace(/\\r\\n/g, "\\n");
{II}if (expected !== got) {{
{III}throw new Error(
{IIII}`Expected the error(s):\\n${{JSON.stringify(expected)}}\\n, ` +
{IIII}`but got:\\n${{JSON.stringify(got)}}\\n` +
{IIII}`when verifying ${{aPath}}`
{III});
{II}}}
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_typescript = typescript_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        deserialization_function = typescript_naming.function_name(
            Identifier(f"{concrete_cls.name}_from_jsonable")
        )

        blocks.extend(
            [
                Stripped(
                    f"""\
test("{cls_name_typescript} round-trip OK", () => {{
{I}const pths = Array.from(
{II}TestCommon.findFilesBySuffixRecursively(
{III}path.join(
{IIII}TestCommon.TEST_DATA_DIR,
{IIII}"Json",
{IIII}"Expected",
{IIII}{typescript_common.string_literal(cls_name_json)}
{III}),
{III}".json"
{II})
{I});
{I}pths.sort();

{I}for (const pth of pths) {{
{II}const jsonable = TestCommon.readJsonFromFileSync(pth);

{II}const instanceOrError = AasJsonization.{deserialization_function}(
{III}jsonable
{II});
{II}expect(instanceOrError.error).toBeNull();
{II}const instance = instanceOrError.mustValue();

{II}TestCommon.assertNoVerificationErrors(AasVerification.verify(instance), pth);

{II}assertSerializeDeserializeEqualsOriginal(
{III}jsonable,
{III}instance,
{III}pth
{II});
{I}}}
}});"""
                ),
                Stripped(
                    f"""\
test("{cls_name_typescript} deserialization fail", () => {{
{I}for (
{II}const causeDir of
{II}TestCommon.findImmediateSubdirectories(
{III}path.join(
{IIII}TestCommon.TEST_DATA_DIR,
{IIII}"Json",
{IIII}"Unexpected",
{IIII}"Unserializable"
{III})
{II})
{I}) {{
{II}// NOTE (mristin):
{II}// Unlike other SDKs, we can not be really sure what additional properties
{II}// JavaScript might bring about. Therefore, we leave out the tests with
{II}// the validation of additional properties.
{II}if (path.basename(causeDir) == "UnexpectedAdditionalProperty") {{
{III}continue;
{II}}}

{II}const clsDir = path.join(
{III}causeDir,
{III}{typescript_common.string_literal(cls_name_json)}
{II});
{II}if (!fs.existsSync(clsDir)) {{
{III}// NOTE (mristin):
{III}// Some classes indeed lack the invalid examples.
{III}continue;
{II}}}

{II}const pths = Array.from(
{III}TestCommon.findFilesBySuffixRecursively(
{IIII}clsDir,
{IIII}".json"
{III})
{II});
{II}pths.sort();

{II}for (const pth of pths) {{
{III}const jsonable = TestCommon.readJsonFromFileSync(pth);

{III}const instanceOrError = AasJsonization.{deserialization_function}(
{IIII}jsonable
{III});
{III}if (instanceOrError.error === null) {{
{IIII}throw new Error(`Expected a de-serialization error for ${{pth}}, but got none`);
{III}}}

{III}assertDeserializationErrorEqualsExpectedOrRecord(
{IIII}instanceOrError.error,
{IIII}pth
{III});
{II}}}
{I}}}
}});"""
                ),
                Stripped(
                    f"""\
test("{cls_name_typescript} verification fail", () => {{
{I}for (
{II}const causeDir of
{II}TestCommon.findImmediateSubdirectories(
{III}path.join(
{IIII}TestCommon.TEST_DATA_DIR,
{IIII}"Json",
{IIII}"Unexpected",
{IIII}"Invalid"
{III})
{II})
{I}) {{
{II}const clsDir = path.join(
{III}causeDir,
{III}{typescript_common.string_literal(cls_name_json)}
{II});
{II}if (!fs.existsSync(clsDir)) {{
{III}// NOTE (mristin):
{III}// Some classes indeed lack the invalid examples.
{III}continue;
{II}}}

{II}const pths = Array.from(
{III}TestCommon.findFilesBySuffixRecursively(
{IIII}clsDir,
{IIII}".json"
{III})
{II});
{II}pths.sort();

{II}for (const pth of pths) {{
{III}const jsonable = TestCommon.readJsonFromFileSync(pth);

{III}const instanceOrError = AasJsonization.{deserialization_function}(
{IIII}jsonable
{III});
{III}if (instanceOrError.error !== null) {{
{IIII}throw new Error(
{IIIII}`Expected no de-serialization error for ${{pth}}, ` +
{IIIII}`but got: ${{instanceOrError.error.message}}: ${{instanceOrError.error.path}}`
{IIII});
{III}}}

{III}const instance = instanceOrError.mustValue();

{III}const verificationErrors = Array.from(AasVerification.verify(instance));
{III}assertVerificationErrorsEqualExpectedOrRecord(
{IIII}verificationErrors,
{IIII}pth
{III});
{II}}}
{I}}}
}});"""
                ),
            ]
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the functions to trace instances for tests."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.golang import common as golang_common
from aas_core_codegen.golang.common import INDENT as I, INDENT2 as II


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(repo_url: Stripped) -> str:
    """Generate the functions to trace instances for tests."""
    blocks = [
        Stripped(
            """\
package aastesting"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}aasstringification "{repo_url}/stringification"
{I}aastypes "{repo_url}/types"
)"""
        ),
        Stripped(
            f"""\
// Represent `that` instance as a human-readable line of an iteration trace.
func TraceMark(that aastypes.IClass) string {{
{I}modelTypeText := aasstringification.MustModelTypeToString(
{II}that.ModelType(),
{I})
{I}return modelTypeText
}}"""
        ),
        golang_common.WARNING,
    ]  # type: List[Stripped]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the ``XxxOrDefault`` methods."""
import io
from typing import List, Optional

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.cpp import common as cpp_common, naming as cpp_naming
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate implementation to test the ``XxxOrDefault`` methods."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "./common.hpp"
#include "./common_examples.hpp"

#include <{include_prefix_path}/stringification.hpp>
#include <{include_prefix_path}/types.hpp>

#define CATCH_CONFIG_MAIN
#include <catch2/catch.hpp>

namespace aas = {library_namespace};"""
        ),
        Stripped(
            f"""\
const std::filesystem::path& DetermineLogDir() {{
{I}static aas::common::optional<std::filesystem::path> result;
{I}if (!result.has_value()) {{
{II}result = test::common::DetermineTestDataDir() / "XxxOrDefault";
{I}}}

{I}return *result;
}}"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        cls_name = cpp_naming.class_name(cls.name)
        interface_name = cpp_naming.interface_name(cls.name)

        load_max = cpp_naming.function_name(Identifier(f"load_max_{cls.name}"))
        load_min = cpp_naming.function_name(Identifier(f"load_min_{cls.name}"))

        for method in cls.methods:
            if not method.name.endswith("_or_default"):
                continue

            assert method.returns is not None, (
                f"Unexpected no return type "
                f"from the method {method.name!r} of {cls.name!r}"
            )

            method_name = cpp_naming.method_name(method.name)

            serialization_snippet: Optional[Stripped] = None

            primitive_type = intermediate.try_primitive_type(method.returns)
            if primitive_type is intermediate.PrimitiveType.BYTEARRAY:
                serialization_snippet = Stripped(
                    f"""\
const std::string serialized(
{I}aas::stringification::Base64Encode(
{II}instance->{method_name}()
{I})
);"""
                )
            elif primitive_type is intermediate.PrimitiveType.STR:
                # NOTE (mristin):
                # We use std::wstring's so we have to convert to a UTF-8.
                serialization_snippet = Stripped(
                    f"""\
const std::string serialized(
{I}aas::common::WstringToUtf8(
{II}instance->{method_name}()
{I})
);"""
                )
            elif primitive_type is not None:
                serialization_snippet = Stripped(
                    f"""\
const std::string serialized(
{I}std::to_string(
{II}instance->{method_name}()
{I})
);"""
                )
            else:
                if isinstance(
                    method.returns, intermediate.OurTypeAnnotation
                ) and isinstance(method.returns.our_type, intermediate.Enumeration):
                    serialization_snippet = Stripped(
                        f"""\
const std::string serialized(
{I}aas::stringification::to_string(
{II}instance->{method_name}()
{I})
);"""
                    )

            if serialization_snippet is None:
                raise NotImplementedError(
                    "We have not implemented the serialization of the type "
                    f"{method.returns} in the tests. If you see this message, "
                    f"please revisit the testgen code and implement it."
                )

            blocks.append(
                Stripped(
                    f"""\
TEST_CASE("Test {method_name} on a min. {cls_name}") {{
{I}const std::shared_ptr<
{II}aas::types::{interface_name}
{I}> instance(
{II}test::common::examples::{load_min}()
{I});

{I}const std::filesystem::path log_path(
{II}DetermineLogDir()
{III}/ "{cls_name}.{method_name}.min.log"
{I});

{I}{indent_but_first_line(serialization_snippet, I)}

{I}test::common::AssertContentEqualsExpectedOrRecord(
{II}serialized,
{II}log_path
{I});
}}"""
                )
            )

            blocks.append(
                Stripped(
                    f"""\
TEST_CASE("Test {method_name} on a max. {cls_name}") {{
{I}const std::shared_ptr<
{II}aas::types::{interface_name}
{I}> instance(
{II}test::common::examples::{load_max}()
{I});

{I}const std::filesystem::path log_path(
{II}DetermineLogDir()
{III}/ "{cls_name}.{method_name}.max.log"
{I});

{I}{indent_but_first_line(serialization_snippet, I)}

{I}test::common::AssertContentEqualsExpectedOrRecord(
{II}serialized,
{II}log_path
{I});
}}"""
                )
            )

    blocks.append(cpp_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate the test code for the enhancing of instances."""

import io

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate the test code for the enhancing of instances."""
    blocks = [
        Stripped("package enhancing_test"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"testing"
{I}aasenhancing "{repo_url}/enhancing"
{I}aastesting "{repo_url}/aastesting"
{I}aastypes "{repo_url}/types"
)"""
        ),
        Stripped(
            f"""\
type Enhancement struct {{
{I}ID int
}}"""
        ),
        Stripped(
            f"""\
func collectIDsAndAssertTheyAreConsecutiveAndTheirCountEqualsNextID(
{I}t *testing.T,
{I}wrapped aastypes.IClass,
{I}nextID int,
) {{
{I}var ids []int

{I}instanceEnh := aasenhancing.MustUnwrap[*Enhancement](wrapped)
{I}ids = append(ids, instanceEnh.ID)

{I}wrapped.Descend(
{II}func(that aastypes.IClass) (abort bool) {{
{III}enh := aasenhancing.MustUnwrap[*Enhancement](that)
{III}ids = append(ids, enh.ID)
{III}return
{II}}},
{I})

{I}if len(ids) != nextID {{
{II}t.Fatalf("Expected to collect %d IDs, but got: %d", len(ids), nextID)
{II}return
{I}}}

{I}for i, id := range ids {{
{II}if id != i {{
{III}t.Fatalf(
{IIII}"Unexpected ID at index %d (starting from 0); expected %d, got %d",
{IIII}i, i, id,
{III})
{II}}}
{I}}}
}}"""
        ),
    ]

    for concrete_cls in symbol_table.concrete_classes:
        must_load_maximal_name = golang_naming.function_name(
            Identifier(f"must_load_maximal_{concrete_cls.name}")
        )

        test_name = golang_naming.function_name(
            (Identifier(f"test_{concrete_cls.name}_wrapped"))
        )

        blocks.append(
            Stripped(
                f"""\
func {test_name}(t *testing.T) {{
{I}instance := aastesting.{must_load_maximal_name}()

{I}nextID := 0
{I}wrapped := aasenhancing.Wrap[*Enhancement](
{II}instance,
{II}func(that aastypes.IClass) (enh *Enhancement, should bool) {{
{III}enh = &Enhancement{{}}
{III}enh.ID = nextID
{III}should = true

{III}nextID++
{III}return
{II}}},
{I})

{I}if !aastesting.DeepEqual(instance, wrapped) {{
{II}t.Fatalf(
{III}"Deep equality failed between the instance and the wrapped: %v %v",
{III}instance, wrapped,
{II})
{I}}}

{I}collectIDsAndAssertTheyAreConsecutiveAndTheirCountEqualsNextID(
{II}t, wrapped, nextID,
{I})
}}"""
            )
        )

        test_name = golang_naming.function_name(
            (Identifier(f"test_{concrete_cls.name}_nothing_wrapped"))
        )

        blocks.append(
            Stripped(
                f"""\
func {test_name}(t *testing.T) {{
{I}instance := aastesting.{must_load_maximal_name}()

{I}wrapped := aasenhancing.Wrap[*Enhancement](
{II}instance,
{II}func(that aastypes.IClass) (enh *Enhancement, should bool) {{
{III}should = false
{III}return
{II}}},
{I})

{I}if !aastesting.DeepEqual(instance, wrapped) {{
{II}t.Fatalf(
{III}"Deep equality failed between the instance and the wrapped: %v %v",
{III}instance, wrapped,
{II})
{I}}}

{I}// Wrapped should be equal to instance by reference as our enhancement factory
{I}// did not wrap anything.
{I}if wrapped != instance {{
{II}t.Fatalf("Unexpected inequality between %v and %v", wrapped, instance)
{I}}}

{I}wrapped.Descend(func (that aastypes.IClass) (abort bool) {{
{II}_, ok := aasenhancing.Unwrap[*Enhancement](that)
{II}if ok {{
{III}t.Fatalf("Unexpected wrapped descendant: %v", that)
{II}}}
{II}return
{I}}})
}}"""
            )
        )

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())
